/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** configrw.c - IOCTL handlers for getting and setting the local bus
**              space configurations.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "ioctl.h"
#include "device.h"
#include "configrw.h"

OS_STATUS
ADMXRC_ReadConfig(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_READCONFIG* iptr;
  ADMXRC_IOCTLS_READCONFIG* optr;
  os_uint32                 val32;
  unsigned long             index;
  MethodResult              res;
  OS_STATUS                 status;

  iptr = req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  optr = req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  index = iptr->in.index;
  res = pDevCtx->methods.configMem(pDevCtx, FALSE, index, &val32);
  if (res == MethodResult_Success) {
    status = OS_STATUS_SUCCESS;
    optr->out.value = val32;
    req->status.length = sizeof(optr->out);
  } else {
    if (res == MethodResult_InvalidLocation) {
      status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    } else if (res == MethodResult_Timeout) {
      status = OS_MAKE_STATUS(ADMXRC_FAILED);
    } else {
      status = OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
    }
  }

  return status;
}

OS_STATUS
ADMXRC_WriteConfig(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_WRITECONFIG* iptr;
  unsigned long              index;
  os_uint32                  val32;
  MethodResult               res;
  OS_STATUS                  status;

  iptr = req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  if (iptr->in.signature1 != 0x587d3f31 ||
      iptr->in.signature2 != 0x043a749f) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  index = iptr->in.index;
  val32 = iptr->in.value;
  res = pDevCtx->methods.configMem(pDevCtx, TRUE, index, &val32);
  if (res == MethodResult_Success) {
    status = OS_STATUS_SUCCESS;
  } else {
    if (res == MethodResult_InvalidLocation) {
      status = OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
    } else if (res == MethodResult_Timeout) {
      status = OS_MAKE_STATUS(ADMXRC_FAILED);
    } else {
      status = OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
    }
  }

  return OS_STATUS_SUCCESS;
}

