#if !defined(_ADATA_DEVICE_DEFS_H)
#define _ADATA_DEVICE_DEFS_H

/*
** Constants and definitions for the ADM-XRC series driver
*/

#define DRIVER_NAME "admxrc2"

/*
** Maximum number of DMA descriptors that can be created per device
*/
#define ADMXRC_MAX_DMADESC                      (256)

/*
** A fixed number of PLXDESCs are created per DMA channel. These
** descriptors are recycled to perform a "chunk" of a DMA transfer,
** as many times as is necessary to complete the DMA transfer.
**
** Thus a DMA transfer that requires more than MAX_PLXDESC pages
** will be split up into a number of chunks, each requiring no more
** than MAX_PLXDESC descriptors to describe it.
**
** The value chosen here should be a reasonable compromise between
** interrupt efficiency and the desire to use no more than a single
** page of memory to hold all of the PLXDESCs for a particular DMA
** channel.
*/
#define ADMXRC_MAX_PLXDESC_PER_CHANNEL          (32)

/*
** The number of PLX DMA channels
*/
#define ADMXRC_NUM_PLX9080_DMA_CHANNEL          (2)
#define ADNXRC_NUM_PLX9656_DMA_CHANNEL          (2)
#define ADMXRC_NUM_DMA_CHANNEL_MAX              (2)

/*
** The number of PLX spaces (S0, S1) etc.
*/
#define ADMXRC_NUM_PLX9080_SPACE                (2)
#define ADMXRC_NUM_PLX9656_SPACE                (2)
#define ADMXRC_NUM_SPACE_MAX                    (2)

/*
** The maximum number of RAM banks on any ADM-XRC variant
*/
#define ADMXRC_NUM_BANK_MAX                     (6)

#define PCI_VENDOR_PLX                          (0x10B5)
#define PCI_VENDOR_ADATA                        (0x4144)
#define PCI_DEVICE_PLX_PCI9080                  (0x9080)
#define PCI_DEVICE_PLX_PCI9656                  (0x9656)
#define PCI_DEVICE_ADATA_ADMXRC2PL              (0x0043)
#define PCI_DEVICE_ADATA_ADMXRC2P               (0x0044)
#define PCI_SUBVENDOR_ADATA                     (0x4144)
#define PCI_SUBDEVICE_ADATA_ADMXRC              (0x0040)
#define PCI_SUBDEVICE_ADATA_ADMXRCP             (0x0040)
#define PCI_SUBDEVICE_ADATA_ADMXRC2L            (0x0041)
#define PCI_SUBDEVICE_ADATA_ADMXRC2             (0x0042)
#define PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCIX      (0x0043)
#define PCI_SUBDEVICE_ADATA_ADMXRC2P_PCI        (0x0044)
#define PCI_SUBDEVICE_ADATA_ADPWRC2             (0x0045)
#define PCI_SUBDEVICE_ADATA_ADPDRC2             (0x0046)
#define PCI_SUBDEVICE_ADATA_ADMXRC2PL_PCI       (0x0047)
#define PCI_SUBDEVICE_ADATA_ADMXRC2P_PCIX       (0x0048)

/*
** Possible modifications to boards.
**
** Note: specific to each board type
*/
#define ADMXRC_MOD_CLOCKSWAP                    (1 << 0)  /* ADMXRC/ADM-XRCP */

/*
** The number of times to try programming the clock generator
*/
#define ADMXRC_PROGRAM_CLOCK_RETRY_LIMIT        (20)

/*
** Experimental feature:
**
**   0x0 => do not allow write and invalidate cycles on PCI bus
**   0x1 => allow write and invalidate cycles on PCI bus
*/
#define ADMXRC_ALLOW_WRITE_INVALIDATE           (0x1U)

/*
**   0x0 => do not allow driver to use chained-mode DMA
**   0x1 => allow driver to use chained-mode DMA
*/
#define ADMXRC_ALLOW_DMA_CHAINED_MODE           (0x1U)

/*
**   0x0 => don't program PCI9656 registers to workaround errata
**   0x1 => program PCI9656 registers to workaround errata
*/
#define ADMXRC_9656_ERRATA_WORKAROUND           (0x0U)

/*
**   0x0 => don't allow DMA transfers direct from user space buffer
**   0x1 => allow DMA transfers direct from user space buffer
*/
#define ADMXRC_ALLOW_DMA_DIRECT_USER_SPACE      (0x1U)

/*
**   0x0 => don't use PCI read multiple commands for DMA
**   0x1 => use PCI read multiple commands for DMA
*/
#define ADMXRC_USE_PCI_READ_MULTIPLE            (0x1U)

#endif
