/*
** ADM-XRC Series Linux Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** linux_params.c - Declare the module parameters.
**
** All of the recognised module parameters must be declared here, so that they
** can be picked up as driver parameters by the generic section of the driver.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"

static long PCIReadMultiple = 1;
static long PCIWriteAndInvalidate = 1;
static long PCI9656ErrataFix = 1;
static long MaxDmaDesc = 256;
static long MaxPlxDesc = 16;
static long ADMXRCMaxLCLK = 40000000;
static long ADMXRCMinLCLK = 400000;
static long ADMXRCPMaxLCLK = 40000000;
static long ADMXRCPMinLCLK = 400000;
static long ADMXRC2LMaxLCLK = 40000000;
static long ADMXRC2LMinLCLK = 400000;
static long ADMXRC2MaxLCLK = 66000000;
static long ADMXRC2MinLCLK = 400000;
static long ADMXRC2PLMaxLCLK = 80000000;
static long ADMXRC2PLMinLCLK = 6000000;
static long ADMXRC2PLUnlockFlash = 0;

static struct {
  char*           mnemonic;
  DriverParamType type;
  void*           value;
} module_params[] = {
  { "PCIReadMultiple", DriverParamType_uint32, &PCIReadMultiple },
  { "PCIWriteAndInvalidate", DriverParamType_uint32, &PCIWriteAndInvalidate },
  { "PCI9656ErrataFix", DriverParamType_uint32, &PCI9656ErrataFix },
  { "MaxDmaDesc", DriverParamType_uint32, &MaxDmaDesc },
  { "MaxPlxDesc", DriverParamType_uint32, &MaxPlxDesc },
  { "ADMXRCMaxLCLK", DriverParamType_uint32, &ADMXRCMaxLCLK },
  { "ADMXRCMinLCLK", DriverParamType_uint32, &ADMXRCMinLCLK },
  { "ADMXRCPMaxLCLK", DriverParamType_uint32, &ADMXRCPMaxLCLK },
  { "ADMXRCPMinLCLK", DriverParamType_uint32, &ADMXRCPMinLCLK },
  { "ADMXRC2LMaxLCLK", DriverParamType_uint32, &ADMXRC2LMaxLCLK },
  { "ADMXRC2LMinLCLK", DriverParamType_uint32, &ADMXRC2LMinLCLK },
  { "ADMXRC2MaxLCLK", DriverParamType_uint32, &ADMXRC2MaxLCLK },
  { "ADMXRC2MinLCLK", DriverParamType_uint32, &ADMXRC2MinLCLK },
  { "ADMXRC2PLMaxLCLK", DriverParamType_uint32, &ADMXRC2PLMaxLCLK },
  { "ADMXRC2PLMinLCLK", DriverParamType_uint32, &ADMXRC2PLMinLCLK },
  { "ADMXRC2PLUnlockFlash", DriverParamType_uint32, &ADMXRC2PLUnlockFlash },
  { NULL } /* Terminator */
};

MODULE_LICENSE("GPL");

MODULE_PARM(PCIReadMultiple, "l");
MODULE_PARM(PCIWriteAndInvalidate, "l");
MODULE_PARM(PCI9656ErrataFix, "l");
MODULE_PARM(MaxDmaDesc, "l");
MODULE_PARM(MaxPlxDesc, "l");
MODULE_PARM(ADMXRCMaxLCLK, "l");
MODULE_PARM(ADMXRCMinLCLK, "l");
MODULE_PARM(ADMXRCPMaxLCLK, "l");
MODULE_PARM(ADMXRCPMinLCLK, "l");
MODULE_PARM(ADMXRC2LMaxLCLK, "l");
MODULE_PARM(ADMXRC2LMinLCLK, "l");
MODULE_PARM(ADMXRC2MaxLCLK, "l");
MODULE_PARM(ADMXRC2MinLCLK, "l");
MODULE_PARM(ADMXRC2PLMaxLCLK, "l");
MODULE_PARM(ADMXRC2PLMinLCLK, "l");
MODULE_PARM(ADMXRC2PLUnlockFlash, "l");

static os_boolean
strNoCaseCmp(
  const char* s1,
  const char* s2)
{
  char a, b;

  while (1) {
    a = *s1++;
    b = *s2++;
    if (a == b) {
      if (a == '\0') {
	return TRUE;
      }
    } else {
      if (a == '\0' || b == '\0') {
	return FALSE;
      }
    }
  }

}

/*
** -----------------------------------------------------------------------
** Exported routines
** -----------------------------------------------------------------------
*/

os_boolean
os_get_driver_param(
  OS_DRIVER_OBJECT* pDrvObj,
  const char*       mnemonic,
  DriverParamType   type,
  void*             value,
  unsigned long     size)
{
  unsigned int i;
  os_boolean   bFound = FALSE;

  for (i = 0; ; i++) {
    if (module_params[i].mnemonic == NULL) {
      break;
    }

    if (strNoCaseCmp(module_params[i].mnemonic, mnemonic) && module_params[i].type == type) {
      bFound = TRUE;

      switch (type) {
      case DriverParamType_uint32:
	*((os_uint32*) value) = (os_uint32) module_params[i].value;
	break;

      case DriverParamType_int32:
	*((os_int32*) value) = (os_int32) module_params[i].value;
	break;

      default:
	break;
      }
    }
  }

  return bFound;
}
