/*
** ADM-XRC Series Driver
**
** (C) Copright 2001-2004 Alpha Data
**
** methods.c - Model-specific routines.
**
** ----------------------------------------------------------------------------
*/

#include "drvparam.h"
#include "flash.h"
#include "methods.h"

static void
dumpDescriptorsPlxbased(
  DeviceContext* pDevCtx,
  PLXDESC* desc,
  unsigned int channel)
{
  unsigned int i = 0;
  os_uint32    offsetl;
  os_uint32    pci, local, count, link;  

  os_dbgprint(0, ("Dump of PLXDESCs:\n"));
  os_dbgprint(0, ("###### DESC------ DESC-BUS---------- [PCI-------- LOCAL----- COUNT----- LINK------]\n"));
	
  offsetl = sizeof(PLXDESC);
  while (1) {
    pci = os_le32_to_cpu(desc->pci);
    local = os_le32_to_cpu(desc->local);
    count = os_le32_to_cpu(desc->count);
    link = os_le32_to_cpu(desc->link);
 
    os_dbgprint(0, ("%6.6ld 0x%p 0x%8lx [%8.8lx %8.8lx %8.8lx %8.8lx]\n",
		   (unsigned long) i,
		   (void*) desc,
		   (unsigned long) pDevCtx->dma.channel[channel].descriptors.pci[i],
		   (unsigned long) pci,
		   (unsigned long) local,
		   (unsigned long) count,
		   (unsigned long) link));
    if (link & PLXDESCF_ENDCHAIN) {
      break;
    }
    desc++;
    i++;
  }
}

static void
dumpDescriptorsXplbased(
  DeviceContext* pDevCtx,
  XPLDMADESC* desc,
  unsigned int channel)
{
  unsigned int i = 0;
  os_uint32 offsetl;
  os_uint32 pah, pal, la, siz, ndh, ndl;
	
  os_dbgprint(0, ("Dump of DMA linked list:\n"));
  os_dbgprint(0, ("### DESC-VIRT- DESC-BUS-- { PCI--------------- LOCAL----- COUNT----- LINK-------------- }\n"));
	
  offsetl = sizeof(XPLDMADESC);
  while (1) {
    pal = os_le32_to_cpu(desc->PAL);
    pah = os_le32_to_cpu(desc->PAH);
    la = os_le32_to_cpu(desc->LA);
    siz = os_le32_to_cpu(desc->SIZ);
    ndl = os_le32_to_cpu(desc->NDL);
    ndh = os_le32_to_cpu(desc->NDH);
    os_dbgprint(0, ("%3.3ld 0x%p 0x%8.8lx { 0x%8.8lx%8.8lx 0x%8.8lx 0x%8.8lx 0x%8.8lx%8.8lx }\n",
		    (unsigned long) i,
		    (void*) desc,
		    (unsigned long) pDevCtx->dma.channel[channel].descriptors.pci[i],
		    (unsigned long) pah, 
		    (unsigned long) pal,
		    (unsigned long) la,
		    (unsigned long) siz,
		    (unsigned long) ndh,
		    (unsigned long) ndl));
    if (ndl & XPLDMADESC_ENDCHAIN) {
      break;
    }
    desc++;
    i++;
  }
}

/*
** Receive up to 32 bits from a 93x6 EEPROM
*/
static void
receive93x6Bits(
  PLX9080REGS* plx,
  os_uint32*   bits,
  unsigned int len,
  os_uint32*   cntrl)
{
  os_uint32    val32 = 0;
  unsigned int i;
  
  for (i = 0; i < len; i++) {
    val32 <<= 1;
    os_write_register_le32(&plx->CNTRL, *cntrl);
    os_read_register_le32(&plx->CNTRL);
    os_write_register_le32(&plx->CNTRL, *cntrl | 0x01000000U);
    os_read_register_le32(&plx->CNTRL);
    os_write_register_le32(&plx->CNTRL, *cntrl);
    if (os_read_register_le32(&plx->CNTRL) & 0x08000000) {
      val32 |= 0x1U;
    }
  }
  
  *bits = val32;
}

/*
** Send up to 32 bits to a 93x6 EEPROM
**
** With PCI9656, always leaves EEDO pin disabled on exit.
*/
static void
send93x6Bits(
  PLX9080REGS* plx,
  os_boolean   pci9656,
  os_uint32    bits,
  unsigned int len,
  os_uint32*   cntrl)
{
  unsigned int	i;

  if (pci9656) {
    /* Enable driver on EEDO pin */
    *cntrl &= ~0x80000000U;
  }
  os_write_register_le32(&plx->CNTRL, *cntrl);
  os_read_register_le32(&plx->CNTRL);
  
  /* Send instruction to EEPROM one bit at a time */
  for (i = 0; i < len; i++) {
    /* Pre-clear the EEPROM data bit */
    *cntrl &= ~0x04000000U;
    /* Set the EEPROM data bit if necessary */
    *cntrl |= (bits & (0x1U << (len - 1))) ? 0x04000000: 0x00000000;
    
    /* Clock the data bit to the EEPROM */
    os_write_register_le32(&plx->CNTRL, *cntrl);
    os_read_register_le32(&plx->CNTRL);
    os_write_register_le32(&plx->CNTRL, *cntrl | 0x01000000);
    os_read_register_le32(&plx->CNTRL);
    os_write_register_le32(&plx->CNTRL, *cntrl);
    os_read_register_le32(&plx->CNTRL);
    
    bits <<= 1;
  }
  
  if (pci9656) {
    /* Disable driver on EEDO pin */
    *cntrl |= 0x80000000U;
    os_write_register_le32(&plx->CNTRL, *cntrl);
    os_read_register_le32(&plx->CNTRL);
  }
}

/*
** Send a command to a 93x6 EEPROM
*/
static void
send93x6Command(
  PLX9080REGS* plx,
  os_boolean   pci9656,
  os_uint32    instr,
  unsigned int len,
  os_uint32*   cntrl)
{	
  /* Chip select asserted */
  *cntrl |=  0x02000000U;
  /* Clear instruction and clock bits */
  *cntrl &= ~0x05000000U;
  os_write_register_le32(&plx->CNTRL, *cntrl);
  os_read_register_le32(&plx->CNTRL);

  /*
  ** Toggle EEPROM's chip select (to get it out of Shift Register Mode)
  ** and assert the chip select.
  */
  os_write_register_le32(&plx->CNTRL, *cntrl & ~0x02000000U);
  os_read_register_le32(&plx->CNTRL);
  os_write_register_le32(&plx->CNTRL, *cntrl);
  os_read_register_le32(&plx->CNTRL);
  
  /* Send the instruction to the EEPROM */
  send93x6Bits(plx, pci9656, instr, len, cntrl);
}

/*
** -----------------------------------------------------------------------
** The error method; should never be called unless something goes wrong
** -----------------------------------------------------------------------
*/
MethodResult
methodError(
  DeviceContext* pDevCtx,
  ...)
{
  os_dbgprint(0, ("*** methodError: called, pDevCtx=0x%p\n", pDevCtx));

  return MethodResult_GeneralFailure;
}

/*
** -----------------------------------------------------------------------
** MethodFlash
** -----------------------------------------------------------------------
*/

MethodResult
methodFlashPlxbased(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  os_uint8  page = (os_uint8) (location >> 21);
  os_uint8* flash = ((os_uint8*) pDevCtx->hw.space1.plxbased) + 0x200000U;

  os_dbgprint(9, ("methodFlashPlxbased: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  if (pDevCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADMXRC2 ||
      pDevCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADPWRC2 ||
      pDevCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADPDRC2) {
    /* Set Flash page register on ADM-XRCII/ADP-WRCII/ADP-DRCII only */
    os_write_register_le8(&pDevCtx->hw.space1.plxbased->variant.admxrc2.FLPAGE, page & 0x7fU);
  } else {
    if (page) {
      return MethodResult_InvalidLocation;
    }
  }
  
  location &= 0x1fffffU;
  if (write) {
    os_write_register_le8(flash + location, (os_uint8) *value);
  } else {
    *value = (os_uint16) os_read_register_le8(flash + location);
  }
  
  return MethodResult_Success;
}

MethodResult
methodFlashXrc2l(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  return MethodResult_NotPresent;
}

MethodResult
methodFlashXpl(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  XPLSPACE1REGS* regs;
  os_uint32      flctl, fldata;
  
  os_dbgprint(9, ("methodFlashXpl: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  regs = pDevCtx->hw.space1.xpl;
  
  /* Drive fl_a, fl_sel_n, fl_oe_n, fl_we_n, assert fl_sel_n */
  flctl = (location & 0x01ffffffU) | (0x1fU << 25);
  os_write_register_le32(&regs->FLCTL, flctl);
  os_dbgprint(10, ("FLCTL=0x%lx\n", (unsigned long) os_read_register_le32(&regs->FLCTL)));
  
  if (write) {
    /* Drive fl_d (data to Flash) */
    fldata = (*value & 0xffU) | 0x100U;
    os_write_register_le32(&regs->FLDATA, fldata);
    /* Assert fl_we_n (write enable for Flash) */
    flctl |= (0x1U << 31);
    os_write_register_le32(&regs->FLCTL, flctl);
    /* Deassert fl_we_n */
    flctl &= ~(0x1U << 31);
    os_write_register_le32(&regs->FLCTL, flctl);
    /* Stop driving fl_d (data to Flash) */
    fldata &= ~0x100U;
    os_write_register_le32(&regs->FLDATA, fldata);
  } else {
    /* Assert fl_oe_n (output enable for Flash) */
    flctl |= (0x1U << 30);
    os_write_register_le32(&regs->FLCTL, flctl);
    /* Read fl_d (data from Flash) */
    *value = (os_uint8) os_read_register_le32(&regs->FLDATA);
    /* Deassert fl_oe_n */
    flctl &= ~(0x1U << 30);
    os_write_register_le32(&regs->FLCTL, flctl);
  }
  
  /* Stop driving fl_a, fl_sel_n, fl_oe_n, fl_we_n, assert fl_sel_n */
  flctl &= ~(0x1fU << 25);
  os_write_register_le32(&regs->FLCTL, flctl);
  
  return MethodResult_Success;
}

MethodResult
methodFlashXrc2n(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  CONTROLWRC2* regs;
  
  os_dbgprint(9, ("methodFlashXrc2n: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  regs = pDevCtx->hw.space1.wrc2;
  
  /* Set up the address */
  os_write_register_le32(&regs->FLADDR, location);
  os_read_register_le32(&regs->FLADDR);
  
  if (write) {
    /* Write the value */
    os_write_register_le8(&regs->FLDATA, (os_uint8) *value);
  } else {
    /* Read the value */
    *value = (os_uint16) os_read_register_le8(&regs->FLDATA);
  }
  
  return MethodResult_Success;
}

MethodResult
methodFlashXprev0(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  return MethodResult_NotPresent;
}

MethodResult
methodFlashXprev1(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  XPSPACE1REGS* regs;
  os_uint32 flctl, fldata;
  os_uint32 val32;
  
  os_dbgprint(9, ("methodFlashXprev1: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  regs = pDevCtx->hw.space1.xp;
  
  fldata = pDevCtx->shadow.xp.fldata;
  
  /* Drive fl_a, fl_sel_n, fl_oe_n, fl_we_n, assert fl_sel_n */
  flctl = ((location & 0x1fffffeU) >> 1) | (0x1fU << 25);
  os_write_register_le32(&regs->FLCTL, flctl);
  val32 = os_read_register_le32(&regs->FLCTL);
  
  if (write) {
    /* Drive fl_d */
    fldata |= (*value | 0x10000U);
    os_write_register_le32(&regs->FLDATA, fldata);
    val32 = os_read_register_le32(&regs->FLDATA);
    
    /* Assert fl_we_n (write enable for Flash) */
    flctl |= (0x1U << 31);
    os_write_register_le32(&regs->FLCTL, flctl);
    val32 = os_read_register_le32(&regs->FLCTL);
    
    /* Deassert fl_we_n */
    flctl &= ~(0x1U << 31);
    os_write_register_le32(&regs->FLCTL, flctl);
    val32 = os_read_register_le32(&regs->FLCTL);
    
    /* Deassert fl_sel_n */
    flctl &= ~(0x1U << 29);
    os_write_register_le32(&regs->FLCTL, flctl);
    val32 = os_read_register_le32(&regs->FLCTL);
  } else {
    /* Assert fl_oe_n */
    flctl |= (0x1U << 30);
    os_write_register_le32(&regs->FLCTL, flctl);
    
    /* Read fl_d */
    *value = (os_uint16) os_read_register_le32(&regs->FLDATA);
    
    /* Deassert fl_oe_n */
    flctl &= ~(0x1U << 30);
    os_write_register_le32(&regs->FLCTL, flctl);
    val32 = os_read_register_le32(&regs->FLCTL);
    
    /* Deassert fl_sel_n */
    flctl &= ~(0x1U << 29);
    os_write_register_le32(&regs->FLCTL, flctl);
    val32 = os_read_register_le32(&regs->FLCTL);
    
    os_dbgprint(9, ("methodFlashXprev1: *value=0x%lx\n",
		    (unsigned long) *value));
  }
  
  /* Stop driving fl_a, fl_sel_n, fl_oe_n, fl_we_n */
  os_write_register_le32(&regs->FLCTL, 0);
  os_read_register_le32(&regs->FLCTL);
  
  /* Stop driving fl_d */
  fldata &= ~0x10000U;
  os_write_register_le32(&regs->FLDATA, fldata);
  os_read_register_le32(&regs->FLDATA);
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** Flash Reset Methods
** -----------------------------------------------------------------------
*/

MethodResult
methodFlashResetPlxbased(
  DeviceContext* pDevCtx)
{
  os_dbgprint(9, ("methodFlashResetPlxbased:"));

  return MethodResult_Success;
}

MethodResult
methodFlashResetXpl(
  DeviceContext* pDevCtx)
{
  os_uint32 fldata;
  XPLSPACE1REGS* regs;
  
  os_dbgprint(9, ("methodFlashResetXpl:"));
  
  regs = pDevCtx->hw.space1.xpl;
  
  /* Drive fl_rst_n, assert fl_rst_n */
  fldata = (0x3U << 30);
  os_write_register_le32(&regs->FLDATA, fldata);
  os_read_register_le32(&regs->FLDATA);
  
  /* Deassert fl_rst_n */
  fldata = (0x1U << 30);
  os_write_register_le32(&regs->FLDATA, fldata);
  os_read_register_le32(&regs->FLDATA);

  /* Stop driving fl_rst_n */
  fldata = (0x0U << 30);
  os_write_register_le32(&regs->FLDATA, fldata);
  os_read_register_le32(&regs->FLDATA);

  /* Wait 20us or so for recovery from reset */
  os_delay_busy_us(20);
  
  return MethodResult_Success;
}

MethodResult
methodFlashResetXp(
  DeviceContext* pDevCtx)
{
  os_uint32 fldata;
  XPSPACE1REGS* regs;
  
  os_dbgprint(9, ("methodFlashResetXp:"));
  
  regs = pDevCtx->hw.space1.xp;
  
  fldata = pDevCtx->shadow.xp.fldata;
  
  /* Drive fl_rst_n, assert fl_rst_n */
  fldata |= (0x3U << 30);
  os_write_register_le32(&regs->FLDATA, fldata);
  os_read_register_le32(&regs->FLDATA);
  
  /* Deassert fl_rst_n */
  fldata &= ~(0x1U << 31);
  os_write_register_le32(&regs->FLDATA, fldata);
  os_read_register_le32(&regs->FLDATA);
  
  /* Stop driving fl_rst_n */
  fldata &= ~(0x1U << 30);
  os_write_register_le32(&regs->FLDATA, fldata);
  os_read_register_le32(&regs->FLDATA);
  
  /* Wait 20us or so for recovery from reset */
  os_delay_busy_us(20);
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodConfigMem
** -----------------------------------------------------------------------
*/

MethodResult
methodConfigMemPlxbased(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned long  location,
  os_uint32*     data)
{
  MethodResult result = MethodResult_Success;
  PLX9080REGS* plx = pDevCtx->hw.adapter.pci9080;
  os_uint32    val32;
  os_uint32    cntrl;
  unsigned int j;
  unsigned int instr_len, addr_len;
  os_uint16    addr_mask;
  os_boolean   pci9656;
  
  switch (pDevCtx->devDesc.id.pci.subDId) {
  case PCI_SUBDEVICE_ADATA_ADMXRC:
  case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    addr_mask = 0x3fU;
    addr_len = 6;
    instr_len = 9;
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2:
  case PCI_SUBDEVICE_ADATA_ADPWRC2:
  case PCI_SUBDEVICE_ADATA_ADPDRC2:
    addr_mask = 0xffU;
    addr_len = 8;
    instr_len = 11;
    break;

  default:
    os_assert(FALSE);
    return MethodResult_GeneralFailure;
  }
  
  if (location > (0x1U << addr_len)) {
    return MethodResult_InvalidLocation;
  }
  
  pci9656 = (pDevCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADMXRC2 ||
	     pDevCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADPWRC2 ||
	     pDevCtx->devDesc.id.pci.subDId == PCI_SUBDEVICE_ADATA_ADPDRC2);
  
  cntrl = os_read_register_le32(&plx->CNTRL);
  
  if (write) {
    val32 = *data;
    
    /* Enable writing to EEPROM */
    send93x6Command(plx, pci9656, EE_WREN << (addr_len - 2), instr_len, &cntrl);

    /* Send write command to EEPROM */
    send93x6Command(plx, pci9656, (EE_WRITE << addr_len) | (location & addr_mask), instr_len, &cntrl);

    /* Send data word to EEPROM */
    send93x6Bits(plx, pci9656, val32, 16, &cntrl);
		
    /* Activate EEPROM's status mechanism: deselect and reselect */
    os_write_register_le32(&plx->CNTRL, cntrl & ~0x02000000);
    os_read_register_le32(&plx->CNTRL);
    os_write_register_le32(&plx->CNTRL, cntrl);
    os_read_register_le32(&plx->CNTRL);

    /*
    ** Poll chip status until write cycle completes (or times out)
    */
    for (j = 0; j < 20000; j++) {
      if (os_read_register_le32(&plx->CNTRL) & 0x08000000U) {
	result = MethodResult_Timeout;
      }
    }
    
    /* Disable writing to EEPROM */
    send93x6Command(plx, pci9656, EE_WDS << addr_len, instr_len, &cntrl);
  } else {
    /* Send read command to EEPROM */
    send93x6Command(plx, pci9656, (EE_READ << addr_len) | (location & addr_mask), instr_len, &cntrl);
		
    /* Get 16-bit word from EEPROM */
    receive93x6Bits(plx, &val32, 16, &cntrl);
    
    *data = (os_uint16) val32;
  }
  
  return result;
}

MethodResult
methodConfigMemXpl(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned long  location,
  os_uint32*     data)
{
  MethodResult	result;

  if (location >= 0x40000U) {
    return MethodResult_InvalidLocation;
  }
  
  if (write) {
    result = writeFlash(pDevCtx, location * sizeof(os_uint32), sizeof(os_uint32), data);
  } else {
    result = readFlash(pDevCtx, location * sizeof(os_uint32), sizeof(os_uint32), data);
  }
  
  return result;
}

/*
** -----------------------------------------------------------------------
** MethodIntEnab
** -----------------------------------------------------------------------
*/

MethodResult
methodIntEnabPlxbased(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  PLX9080REGS*	regs;
  os_uint32			intcsr;
  os_uint8			imask;
  
  regs = pDevCtx->hw.adapter.pci9080;
  
  if (enable) {
    /* Establish the DMACSR register: both DMA channels initially enabled */
    pDevCtx->shadow.plxbased.dmacsr = 0x0101U;
    os_write_register_le16(&regs->DMACSR.asWord, pDevCtx->shadow.plxbased.dmacsr);
    
    /* Clear any outstanding FPGA interrupts */
    os_write_register_le8(&pDevCtx->hw.space1.plxbased->ICON, 0x1U);
    
    /* Unmask the FPGA interrupt in CPLD*/
    pDevCtx->shadow.plxbased.imask = 0x1U;
    os_write_register_le8(&pDevCtx->hw.space1.plxbased->IMSET, 0xffU);
    os_write_register_le8(&pDevCtx->hw.space1.plxbased->IMCLR, pDevCtx->shadow.plxbased.imask);
    
    /* Enable PCI9080/PCI9656 interrupts */
    pDevCtx->shadow.plxbased.intcsr =
      PLX_INTCSR_PCI_IE |
      PLX_INTCSR_LOC_IE |
      PLX_INTCSR_DMA0_IE |
      PLX_INTCSR_DMA1_IE;
    pDevCtx->shadow.plxbased.intcsr_mask =
      PLX_INTCSR_LOC_INT |
      PLX_INTCSR_DMA0_INT |
      PLX_INTCSR_DMA1_INT;
#if 0
    /* Driver does not dismiss these interrupts as yet, so ignore them */
    pDevCtx->shadow.plxbased.intcsr_mask |=
      PLX_INTCSR_DM_ABT_INT |
      PLX_INTCSR_DMA0_ABT_INT |
      PLX_INTCSR_DMA1_ABT_INT |
      PLX_INTCSR_TGT_ABT_INT;
#endif
    os_write_register_le32(&regs->INTCSR, pDevCtx->shadow.plxbased.intcsr);
    
    intcsr = os_read_register_le32(&regs->INTCSR);
    imask = os_read_register_le8(&pDevCtx->hw.space1.plxbased->IMSET);
    os_dbgprint(3, ("methodIntEnabPlxbased: interrupts enabled, PLX.INTCSR=0x%8.8lx CPLD.IMASK=0x%2.2lx intcsr_mask=0x%8.8lx\n",
		    (unsigned long) intcsr, (unsigned long) imask,
		    (unsigned long) pDevCtx->shadow.plxbased.intcsr_mask));
  } else {
    /* Disable any interrupts from the PCI9080/PCI9656 */
    pDevCtx->shadow.plxbased.intcsr = 0x0U;
    pDevCtx->shadow.plxbased.intcsr_mask = 0x0U;
    os_write_register_le32(&regs->INTCSR, pDevCtx->shadow.plxbased.intcsr);
    
    /* Mask the FPGA interrupt in the CPLD */
    pDevCtx->shadow.plxbased.imask = 0x0U;
    os_write_register_le8(&pDevCtx->hw.space1.plxbased->IMSET, 0x1U);
    
    intcsr = os_read_register_le32(&regs->INTCSR);
    imask = os_read_register_le8(&pDevCtx->hw.space1.plxbased->IMSET);
    os_dbgprint(3, ("methodIntEnabPlxbased: interrupts disabled, PLX.INTCSR=0x%8.8lx CPLD.IMASK=0x%2.2lx\n",
		    (unsigned long) intcsr, (unsigned long) imask));
  }
  
  return MethodResult_Success;
}

MethodResult
methodIntEnabXplrev0(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  XPLBRIDGEREGS* regs;
  XPLSPACE1REGS* s1regs;
  os_uint32      pictl, fictl;

  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.space1.xpl;
  
  if (enable) {
    /* Clear FPGA interrupt */
    os_write_register_le32(&s1regs->FISTAT, 0x1U);
    
    /* Enable FPGA interrupt */
    pDevCtx->shadow.xpl.fictl = 0x1U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.xpl.fictl);
    
    /* Enable bridge interrupts */
    pDevCtx->shadow.xpl.pictl = 0x80000001U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.xpl.pictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXplrev0: interrupts enabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  } else {
    /* Disable bridge interrupts */
    pDevCtx->shadow.xpl.pictl = 0x0U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.xpl.pictl);
    
    /* Disable FPGA interrupt */
    pDevCtx->shadow.xpl.fictl = 0x0U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.xpl.fictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXplrev0: interrupts disabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  }

  return MethodResult_Success;
}

MethodResult
methodIntEnabXplrev1(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  XPLBRIDGEREGS* regs;
  XPLSPACE1REGS* s1regs;
  os_uint32 pictl, fictl;
  unsigned int	i;

  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.space1.xpl;

  if (enable) {
    /* Clear FPGA interrupt */
    os_write_register_le32(&s1regs->FISTAT, 0x1U);
		
    /* Enable FPGA interrupt */
    pDevCtx->shadow.xpl.fictl = 0x1U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.xpl.fictl);
    
    /* Enable bridge interrupts */
    pDevCtx->shadow.xpl.pictl = 0x8000ff01U;
    os_write_register_le32(&regs->PISTAT, 0xffffffff);
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.xpl.pictl);
    
    /* Initialise DMA channel state */
    for (i = 0; i < NUM_XPLBRIDGE_DMA_CHANNEL; i++) {
      pDevCtx->shadow.xpl.dxctl[i] = 0x4U;
      os_write_register_le32(&regs->DMA[i].CTL, pDevCtx->shadow.xpl.dxctl[i]);
    }
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXplrev0: interrupts enabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
    
  } else {
    /* Disable bridge interrupts */
    pDevCtx->shadow.xpl.pictl = 0x0U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.xpl.pictl);
    
    /* Disable FPGA interrupt */
    pDevCtx->shadow.xpl.fictl = 0x0U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.xpl.fictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXplrev0: interrupts disabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		   (unsigned long) pictl, (unsigned long) fictl));
  }
  
  return MethodResult_Success;
}

MethodResult
methodIntEnabXp(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  XPLBRIDGEREGS* regs;
  XPSPACE1REGS* s1regs;
  os_uint32 pictl, fictl, fldata, dxctl[2];
  unsigned int	i;
  
  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.space1.xp;
  
  os_dbgprint(1, ("methodIntEnabXp: entered\n"));
  
  if (enable) {
    /* Clear FPGA interrupt */
    os_dbgprint(2, ("methodIntEnabXp: clearing FPGA interrupt\n"));
    os_write_register_le32(&s1regs->FISTAT, 0x1U);
    
    /* Enable FPGA interrupt */
    os_dbgprint(2, ("methodIntEnabXp: enabling FPGA interrupt\n"));
    pDevCtx->shadow.xp.fictl = 0x1U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.xp.fictl);
    
    /* Enable bridge interrupts */
    os_dbgprint(2, ("methodIntEnabXp: enabling interrupts in bridge\n"));
    pDevCtx->shadow.xp.pictl = 0x8000ff01U;
    os_write_register_le32(&regs->PISTAT, 0xffffffff);
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.xp.pictl);
    
    /* Initialise DMA channel state */
    os_dbgprint(2, ("methodIntEnabXp: initializing DMA channel state\n"));
    for (i = 0; i < NUM_XPLBRIDGE_DMA_CHANNEL; i++) {
      pDevCtx->shadow.xp.dxctl[i] = 0x4U;
      os_write_register_le32(&regs->DMA[i].CTL, pDevCtx->shadow.xp.dxctl[i]);
      os_read_register_le32(&regs->DMA[i].CTL);
    }
    
    /* Initialise Flash access stuff */
    pDevCtx->shadow.xp.fldata = (0x3U << 26);
    os_write_register_le32(&s1regs->FLDATA, pDevCtx->shadow.xp.fldata);
    os_read_register_le32(&s1regs->FLDATA);
    
    pictl = os_read_register_le32(&regs->PICTL);
    dxctl[0] = os_read_register_le32(&regs->DMA[0].CTL);
    dxctl[1] = os_read_register_le32(&regs->DMA[1].CTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    fldata = os_read_register_le32(&s1regs->FLDATA);
    os_dbgprint(1, ("methodIntEnabXp: interrupts enabled\n"));
    os_dbgprint(1, ("methodIntEnabXp: Bridge.D0CTL=0x%8.8lx Bridge.D1CTL=0x%8.8lx\n",
		    (unsigned long) dxctl[0], (unsigned long) dxctl[1]));
    os_dbgprint(1, ("methodIntEnabXp: Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx S1.FLDATA=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl, (unsigned long) fldata));
  } else {
    /* Disable bridge interrupts */
    pDevCtx->shadow.xp.pictl = 0x0U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.xp.pictl);
    
    /* Disable FPGA interrupt */
    pDevCtx->shadow.xpl.fictl = 0x0U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.xp.fictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXp: interrupts disabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  }
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodIsr
** -----------------------------------------------------------------------
*/

os_boolean
methodIsrPlxbased(
  DeviceContext* pDevCtx)
{
  os_uint32  intcsr;
  os_uint8   icon;
  os_uint16  dmacsr;
  os_boolean bHandled = FALSE;

  intcsr = os_read_register_le32(&pDevCtx->hw.adapter.pci9080->INTCSR) & pDevCtx->shadow.plxbased.intcsr_mask;
	
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (intcsr & PLX_INTCSR_LOC_INT) {
    icon = os_read_register_8(&pDevCtx->hw.space1.plxbased->ICON) & pDevCtx->shadow.plxbased.imask;

    if (icon) {
      os_dbgprint(6, ("methodIsrPlxbased: FPGA interrupt\n"));

      /* Clear the FPGA interrupt in the CPLD */
      os_write_register_8(&pDevCtx->hw.space1.plxbased->ICON, icon);

      /* Schedule the DPC that does the real work of handling the FPGA interrupt */
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
    }
    bHandled = TRUE;
  }

  /* Is there a PCI9080/PCI9656 DMA interrupt? */
  if (intcsr & (PLX_INTCSR_DMA0_INT | PLX_INTCSR_DMA1_INT)) {
    os_dbgprint(6, ("methodIsrPlxbased: DMA interrupt\n"));
    
    /* Find out which DMA channels are interrupting */
    dmacsr = 0x0U;
    if (intcsr & PLX_INTCSR_DMA0_INT) {
      dmacsr |= 0x0008U;
    }
    if (intcsr & PLX_INTCSR_DMA1_INT) {
      dmacsr |= 0x0800U;
    }
    
    if (dmacsr) {
      /* Clear the pending DMA interrupts(s) */
      os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord,
			     (os_uint16) (dmacsr | pDevCtx->shadow.plxbased.dmacsr));  

      /* Schedule the DPC that does the real work of handling the DMA interrupt */
      if (dmacsr & 0x0008U) {
	pDevCtx->dma.channel[0].completionCode = DmaCompletionCode_Normal;
	os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[0], (void*) 0);
      }
      if (dmacsr & 0x0800U) {
	pDevCtx->dma.channel[1].completionCode = DmaCompletionCode_Normal;
	os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[1], (void*) 1);
      }
    }
    
    bHandled = TRUE;
  }

#if 0
  /* Is there a direct master PCI abort interrupt? */
  if (!(intcsr & PLX_INTCSR_DM_ABT_INT)) {
    os_dbgprint(0, ("*** methodIsrPlxbased: Direct master abort interrupt\n"));
  }

  /* Is there a DMA channel 0 PCI abort interrupt? */
  if (!(intcsr & PLX_INTCSR_DMA0_ABT_INT)) {
    os_dbgprint(0, ("*** methodIsrPlxbased: DMA channel 0 abort interrupt\n"));
  }

  /* Is there a DMA channel 1 PCI abort interrupt? */
  if (!(intcsr & PLX_INTCSR_DMA1_ABT_INT)) {
    os_dbgprint(0, ("*** methodIsrPlxbased: DMA channel 1 abort interrupt\n"));
  }

  /* Is there a PCI target abort interrupt? */
  if (!(intcsr & PLX_INTCSR_TGT_ABT_INT)) {
    os_dbgprint(0, ("*** methodIsrPlxbased: target abort interrupt\n"));
  }
#endif

#if 0
  if (!bHandled) {
    /* Not necessarily an error - we may be called if we're sharing an interrupt */
    os_dbgprint(0, ("*** methodIsrPlxbased: unhandled interrupt, intcsr=0x%8.8lx intcsr_mask=0x%8.8lx\n",
		    (unsigned long) intcsr, (unsigned long) pDevCtx->shadow.plxbased.intcsr_mask));
  }
#endif

  return bHandled;
}

os_boolean
methodIsrXpl(
  DeviceContext* pDevCtx)
{
  os_boolean bHandled = FALSE;
  os_uint32 pistat, fistat;
  XPLBRIDGEREGS* regs;
  XPLSPACE1REGS* s1regs;
  unsigned int i;

  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.space1.xpl;

  /* Get the pending interrupts */
  pistat = os_read_register_le32(&regs->PISTAT) & pDevCtx->shadow.xpl.pictl;

  /* Clear the pending interrupts */
  os_write_register_le32(&regs->PISTAT, pistat);
  
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (pistat & 0x1U) {
    fistat = os_read_register_le32(&s1regs->FISTAT) & pDevCtx->shadow.xpl.fictl;
   
    /* Clear the pending interrupts */
    os_write_register_le32(&s1regs->FISTAT, fistat);
    
    if (fistat & 0x1U) {
      os_dbgprint(6, ("methodIsrXpl: FPGA interrupt\n"));
      
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
      bHandled = TRUE;
    }
  }

  /* Are any DMA channels interrupting? */
  for (i = 0; i < 8; i++) {
    if (pistat & (0x100 << i)) {
      os_uint32 dxctl;
      os_uint8 code;
      
      os_dbgprint(6, ("methodIsrXp: DMA channel %ld interrupt\n",
		      (unsigned long) i));
      dxctl = (os_read_register_le32(&regs->DMA[i].CTL) >> 6) & 0x3fU;
      if (!dxctl) {
	code = DmaCompletionCode_Normal;
      } else if (dxctl & 0x37U) {
	code = DmaCompletionCode_Failed;
      } else {
	code = DmaCompletionCode_EotRequested;
      }
      pDevCtx->dma.channel[i].completionCode = code;
      os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[i], (void*) i);
      bHandled = TRUE;
    }
  }
  
  return bHandled;
}

os_boolean
methodIsrXp(
  DeviceContext* pDevCtx)
{
  os_boolean bHandled = FALSE;
  os_uint32 pistat, fistat;
  XPLBRIDGEREGS* regs;
  XPSPACE1REGS* s1regs;
  unsigned int i;

  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.space1.xp;
  
  /* Get the pending interrupts */
  pistat = os_read_register_le32(&regs->PISTAT) & pDevCtx->shadow.xp.pictl;
  
  /* Clear the pending interrupts */
  os_write_register_le32(&regs->PISTAT, pistat);
  
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (pistat & 0x1U) {
    fistat = os_read_register_le32(&s1regs->FISTAT) & pDevCtx->shadow.xp.fictl;
    
    /* Clear the pending interrupts */
    os_write_register_le32(&s1regs->FISTAT, fistat);
    
    if (fistat & 0x1U) {
      os_dbgprint(6, ("methodIsrXp: FPGA interrupt\n"));
      
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
      bHandled = TRUE;
    }
  }
  
  /* Are any DMA channels interrupting? */
  for (i = 0; i < 8; i++) {
    if (pistat & (0x100 << i)) {
      os_uint32 dxctl;
      os_uint8 code;
      
      os_dbgprint(6, ("methodIsrXp: DMA channel %ld interrupt\n",
		      (unsigned long) i));
      dxctl = (os_read_register_le32(&regs->DMA[i].CTL) >> 6) & 0x3fU;
      if (!dxctl) {
	code = DmaCompletionCode_Normal;
      } else if (dxctl & 0x37U) {
	code = DmaCompletionCode_Failed;
      } else {
	code = DmaCompletionCode_EotRequested;
      }
      pDevCtx->dma.channel[i].completionCode = code;
      os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[i], (void*) i);
      bHandled = TRUE;
    }
  }
  
  return bHandled;
}

/*
** -----------------------------------------------------------------------
** MethodFpga
** -----------------------------------------------------------------------
*/

MethodResult
methodFpgaPlxbased(
  DeviceContext* pDevCtx,
  MethodFpgaOp   op,
  void*          operand)
{
  os_uint8	fcon;
  
  switch (op) {
  case MethodFpgaOp_SetProg:
    {
      os_boolean	asserted = *((os_boolean*) operand);
      
      fcon = os_read_register_le8(&pDevCtx->hw.space1.plxbased->FCON) & ~0x6U;
      if (asserted) {
	fcon = 0x1U;
      } else {
	fcon &= ~0x1U;
      }
      os_write_register_le8(&pDevCtx->hw.space1.plxbased->FCON, fcon);
      break;
    }
    
  case MethodFpgaOp_GetInit:
    {
      fcon = os_read_register_le8(&pDevCtx->hw.space1.plxbased->FCON);
      *((os_boolean*) operand) = (fcon & 0x2U) != 0;
      break;
    }
    
  case MethodFpgaOp_GetDone:
    {
      fcon = os_read_register_le8(&pDevCtx->hw.space1.plxbased->FCON);
      *((os_boolean*) operand) = (fcon & 0x4U) != 0;
      break;
    }
    
  default:
    return MethodResult_NotSupported;
  }
  
  return MethodResult_Success;
}

MethodResult
methodFpgaXpl(
  DeviceContext* pDevCtx,
  MethodFpgaOp   op,
  void*          operand)
{
  os_uint32 fpctl;

  switch (op) {
  case MethodFpgaOp_SetProg:
    {
      os_boolean	asserted = *((os_boolean*) operand);
      
      fpctl = os_read_register_le32(&pDevCtx->hw.space1.xpl->FPCTL) & ~0x6U;
      if (asserted) {
	fpctl |= 0x1U;
      } else {
	fpctl &= ~0x1U;
      }
      os_write_register_le32(&pDevCtx->hw.space1.xpl->FPCTL, fpctl);
      break;
    }
    
  case MethodFpgaOp_GetInit:
    {
      fpctl = os_read_register_le32(&pDevCtx->hw.space1.xpl->FPCTL);
      *((os_boolean*) operand) = (fpctl & 0x2U) != 0;
      break;
    }
    
  case MethodFpgaOp_GetDone:
    {
      fpctl = os_read_register_le32(&pDevCtx->hw.space1.xpl->FPCTL);
      *((os_boolean*) operand) = (fpctl & 0x4U) != 0;
      break;
    }
    
  default:
    return MethodResult_NotSupported;
  }
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodSelectMap
** -----------------------------------------------------------------------
*/

MethodResult
methodSelectMapPlxbased(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint8*      value)
{
  if (write) {
    os_write_register_le8(&pDevCtx->hw.space1.plxbased->SELECTMAP[0], *value);
  } else {
    *value = os_read_register_le8(&pDevCtx->hw.space1.plxbased->SELECTMAP[0]);
  }
  
  return MethodResult_Success;
}

MethodResult
methodSelectMapXpl(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint8*      value)
{
  if (write) {
    os_write_register_le8(&pDevCtx->hw.space1.xpl->FPSMAP[1], (os_uint8) *value);
  } else {
    *value = (os_uint8) (os_read_register_le8(&pDevCtx->hw.space1.xpl->FPSMAP[1]));
  }
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodConfigSpace
** -----------------------------------------------------------------------
*/

MethodResult
methodConfigSpacePlxbased(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned int   space,
  os_uint32*     value)
{
  os_uint32 lbrd0;
  os_uint32 f;

  if (space > 1) {
    return MethodResult_NotSupported;
  }
  
  if (write) {
    f = *value;
    lbrd0 = os_read_register_le32(&pDevCtx->hw.adapter.pci9080->LBRD0);
    
    if (f & ADMXRC2_SPACE_SET_WIDTH) {
      if ((f & ADMXRC2_SPACE_MASK_WIDTH) == ADMXRC2_SPACE_WIDTH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | ADMXRC2_SPACE_WIDTH_32;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_WIDTH);
    }
    
    if (f & ADMXRC2_SPACE_SET_PREFETCH) {
      if ((f & ADMXRC2_SPACE_MASK_PREFETCH) == ADMXRC2_SPACE_PREFETCH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | ADMXRC2_SPACE_PREFETCH_MAXIMUM;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_PREFETCH);
    }
    
    if (f & ADMXRC2_SPACE_SET_BURST) {
      if ((f & ADMXRC2_SPACE_MASK_BURST) == ADMXRC2_SPACE_BURST_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_BURST) | ADMXRC2_SPACE_BURST_ENABLED;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_BURST) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_BURST);
    }
    
    f &= ~(ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_SET_BURST | ADMXRC2_SPACE_SET_PREFETCH);
    
    switch (f & ADMXRC2_SPACE_MASK_WIDTH) {
    case ADMXRC2_SPACE_WIDTH_8:
      lbrd0 = (lbrd0 & ~0x3U) | 0x0U;
      break;
      
    case ADMXRC2_SPACE_WIDTH_16:
      lbrd0 = (lbrd0 & ~0x3U) | 0x1U;
      break;
      
    case ADMXRC2_SPACE_WIDTH_32:
      lbrd0 = (lbrd0 & ~0x3U) | 0x2U;
      break;
      
    case ADMXRC2_SPACE_WIDTH_64:
      return MethodResult_NotSupported;
      
    default:
      return MethodResult_InvalidParameter;
    }
    
    switch (f & ADMXRC2_SPACE_MASK_PREFETCH) {
    case ADMXRC2_SPACE_PREFETCH_MINIMUM:
      /* No prefetch */
      lbrd0 = (lbrd0 & ~0x100U) | (0x1U << 8);
      break;
      
    case ADMXRC2_SPACE_PREFETCH_NORMAL:
      /* 64 bytes prefetch */
      lbrd0 = (lbrd0 & ~0x500U) | (0x1U << 10);
      lbrd0 = (lbrd0 & ~0x7800U) | (0x0U << 11);
      break;
      
    case ADMXRC2_SPACE_PREFETCH_MAXIMUM:
      /* Unlimited prefetch */
      lbrd0 = (lbrd0 & ~0x500U) | (0x0U << 10);
      lbrd0 = (lbrd0 & ~0x7800U) | (0x0U << 11);
      break;
      
    default:
      return MethodResult_InvalidParameter;
    }
    
    switch (f & ADMXRC2_SPACE_MASK_BURST) {
    case ADMXRC2_SPACE_BURST_DISABLED:
      lbrd0 = (lbrd0 & ~0x1000000U) | (0x0U << 24);
      break;
      
    case ADMXRC2_SPACE_BURST_ENABLED:
      lbrd0 = (lbrd0 & ~0x1000000U) | (0x1U << 24);
      break;
      
    default:
      return MethodResult_InvalidParameter;
    }
    
    os_write_register_le32(&pDevCtx->hw.adapter.pci9080->LBRD0, lbrd0);
    pDevCtx->info.spaceConfig[space] = f;
  } else {
    *value = pDevCtx->info.spaceConfig[space];
  }
  
  return MethodResult_Success;
}

MethodResult
methodConfigSpaceXplrev0(
  DeviceContext* pDevCtx,
  os_boolean write,
  unsigned int space,
  os_uint32* value)
{
  os_uint32 ls0cfg;
  os_uint32 f = *value;

  if (space > 1) {
    return MethodResult_NotSupported;
  }

  if (write) {
    ls0cfg = os_read_register_le32(&pDevCtx->hw.adapter.xpl->LS0CFG);
		
    if (f & ADMXRC2_SPACE_SET_WIDTH) {
      if ((f & ADMXRC2_SPACE_MASK_WIDTH) == ADMXRC2_SPACE_WIDTH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | ADMXRC2_SPACE_WIDTH_32;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_WIDTH);
    }

    if (f & ADMXRC2_SPACE_SET_PREFETCH) {
      if ((f & ADMXRC2_SPACE_MASK_PREFETCH) == ADMXRC2_SPACE_PREFETCH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | ADMXRC2_SPACE_PREFETCH_NORMAL;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_PREFETCH);
    }

    if (f & ADMXRC2_SPACE_SET_BURST) {
      if ((f & ADMXRC2_SPACE_MASK_BURST) == ADMXRC2_SPACE_BURST_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_BURST) | ADMXRC2_SPACE_BURST_ENABLED;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_BURST) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_BURST);
    }

    f &= ~(ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_SET_BURST | ADMXRC2_SPACE_SET_PREFETCH);

    switch (f & ADMXRC2_SPACE_MASK_WIDTH) {
    case ADMXRC2_SPACE_WIDTH_8:
    case ADMXRC2_SPACE_WIDTH_16:
      return MethodResult_NotSupported;
			
    case ADMXRC2_SPACE_WIDTH_32:
      ls0cfg = (ls0cfg & ~0x3U) | 0x2U;
      break;
			
    case ADMXRC2_SPACE_WIDTH_64:
      ls0cfg = (ls0cfg & ~0x3U) | 0x3U;
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    switch (f & ADMXRC2_SPACE_MASK_PREFETCH) {
    case ADMXRC2_SPACE_PREFETCH_MINIMUM:
      /* 8 bytes prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x0U << 5);
      break;
			
    case ADMXRC2_SPACE_PREFETCH_NORMAL:
      /* 64 bytes prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x1U << 5) | (0x2U << 2);
      break;
			
    case ADMXRC2_SPACE_PREFETCH_MAXIMUM:
      /* Unlimited prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x1U << 5) | (0x4U << 2);
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    switch (f & ADMXRC2_SPACE_MASK_BURST) {
    case ADMXRC2_SPACE_BURST_DISABLED:
      return MethodResult_NotSupported;
			
    case ADMXRC2_SPACE_BURST_ENABLED:
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    os_write_register_le32(&pDevCtx->hw.adapter.xpl->LS0CFG, ls0cfg);
    pDevCtx->info.spaceConfig[space] = f;
  } else {
    *value = pDevCtx->info.spaceConfig[space];
  }

  return MethodResult_Success;
}

MethodResult
methodConfigSpaceXplrev1(
  DeviceContext* pDevCtx,
  os_boolean write,
  unsigned int space,
  os_uint32* value)
{
  os_uint32 ls0cfg;
  os_uint32 f = *value;

  if (space > 1) {
    return MethodResult_NotSupported;
  }

  if (write) {
    ls0cfg = os_read_register_le32(&pDevCtx->hw.adapter.xpl->LS0CFG);
		
    if (f & ADMXRC2_SPACE_SET_WIDTH) {
      if ((f & ADMXRC2_SPACE_MASK_WIDTH) == ADMXRC2_SPACE_WIDTH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | ADMXRC2_SPACE_WIDTH_32;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_WIDTH);
    }

    if (f & ADMXRC2_SPACE_SET_PREFETCH) {
      if ((f & ADMXRC2_SPACE_MASK_PREFETCH) == ADMXRC2_SPACE_PREFETCH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | ADMXRC2_SPACE_PREFETCH_NORMAL;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_PREFETCH);
    }

    if (f & ADMXRC2_SPACE_SET_BURST) {
      if ((f & ADMXRC2_SPACE_MASK_BURST) == ADMXRC2_SPACE_BURST_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_BURST) | ADMXRC2_SPACE_BURST_ENABLED;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_BURST) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_BURST);
    }

    f &= ~(ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_SET_BURST | ADMXRC2_SPACE_SET_PREFETCH);

    switch (f & ADMXRC2_SPACE_MASK_WIDTH) {
    case ADMXRC2_SPACE_WIDTH_8:
    case ADMXRC2_SPACE_WIDTH_16:
      return MethodResult_NotSupported;
			
    case ADMXRC2_SPACE_WIDTH_32:
      ls0cfg = (ls0cfg & ~0x3U) | 0x2U;
      break;
			
    case ADMXRC2_SPACE_WIDTH_64:
      ls0cfg = (ls0cfg & ~0x3U) | 0x3U;
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    switch (f & ADMXRC2_SPACE_MASK_PREFETCH) {
    case ADMXRC2_SPACE_PREFETCH_MINIMUM:
      /* 8 bytes prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x0U << 5);
      break;
			
    case ADMXRC2_SPACE_PREFETCH_NORMAL:
      /* 64 bytes prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x1U << 5) | (0x2U << 2);
      break;
			
    case ADMXRC2_SPACE_PREFETCH_MAXIMUM:
      /* Unlimited prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x1U << 5) | (0x4U << 2);
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    switch (f & ADMXRC2_SPACE_MASK_BURST) {
    case ADMXRC2_SPACE_BURST_DISABLED:
      ls0cfg &= ~(0x1U << 6);
      break;
			
    case ADMXRC2_SPACE_BURST_ENABLED:
      ls0cfg |= (0x1U << 6);
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    os_write_register_le32(&pDevCtx->hw.adapter.xpl->LS0CFG, ls0cfg);
    pDevCtx->info.spaceConfig[space] = f;
  } else {
    *value = pDevCtx->info.spaceConfig[space];
  }

  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodAddDmaBuffer
** -----------------------------------------------------------------------
*/

MethodResult
methodAddDmaBufferPlxbased(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  unsigned int    index,
  DmaScatterNode* snode)
{
  PLXDESC*  desc;
  os_uint32 flags;

  /* Set up PLX descriptor */
  desc = &pDevCtx->dma.channel[channel].descriptors.aligned.plxbased[index];
  if (index) {
    (desc - 1)->link |= os_cpu_to_le32(pDevCtx->dma.channel[channel].descriptors.pci[index]);
  }
  desc->pci   = os_cpu_to_le32(snode->busAddr);
  desc->local = os_cpu_to_le32(snode->localAddr);
  desc->count = os_cpu_to_le32(snode->length);
  flags = snode->bWrite ? 0x1U : 0x9U;
  desc->link  = os_cpu_to_le32(flags);

  return MethodResult_Success;
}

MethodResult
methodAddDmaBufferXplrev0(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  unsigned int    index,
  DmaScatterNode* snode)
{
  return MethodResult_NotSupported;
}

MethodResult methodAddDmaBufferXplrev1(
  DeviceContext* pDevCtx,
  unsigned int channel,
  unsigned int index,
  DmaScatterNode* scatterNode)
{
  XPLDMADESC* desc;
  os_uint32 sizFlags;

  desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[index];
  if (index) {
    (desc - 1)->NDL |= os_cpu_to_le32(pDevCtx->dma.channel[channel].descriptors.pci[index]);
  }
  if (scatterNode->bWrite) {
    sizFlags = XPLDMADESC_PCITOLOCAL;
  } else {
    sizFlags = XPLDMADESC_LOCALTOPCI;
  }
  desc->PAL = os_cpu_to_le32((os_uint32) scatterNode->busAddr);
  desc->PAH = os_cpu_to_le32(0);
  desc->LA = os_cpu_to_le32(scatterNode->localAddr);
  desc->SIZ = os_cpu_to_le32(scatterNode->length | sizFlags);
  desc->NDL = os_cpu_to_le32(XPLDMADESC_PCISPACE);
  desc->NDH = os_cpu_to_le32(0);
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodDoDma
** -----------------------------------------------------------------------
*/

MethodResult
methodDoDmaPlxbased(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  os_boolean      start,
  os_uint32       mode,
  unsigned int    numBuffer)
{
  MethodResult res = MethodResult_Success;
  os_uint16    mask, dmacsr, val16;
  PLXDESC*     desc;

  dmacsr = pDevCtx->shadow.plxbased.dmacsr;
  os_dbgprint(6, ("methodDoDmaPlxbased: channel=%ld start=%ld mode=0x%lx, dmacsr_shadow=0x%4.4lx\n",
		  (unsigned long) channel,
		  (unsigned long) start,
		  (unsigned long) mode,
		  (unsigned long) dmacsr));
      
  if (start) {
    /*
    ** Kick off a chunk of a DMA transfer
    */

    desc = &pDevCtx->dma.channel[channel].descriptors.aligned.plxbased[numBuffer - 1];
    desc->link |= os_cpu_to_le32(PLXDESCF_ENDCHAIN | PLXDESCF_INTERRUPT);

    if (os_dbglevel(8)) {
      dumpDescriptorsPlxbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.plxbased,
			      channel);
    }
  
    mode |=
      (0x1U <<  9)  | /* Use chained mode           */
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (driverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    mask = channel ? 0x0200U : 0x0002U;
    val16 = os_read_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord);
    if (val16 & (mask << 3)) {
      /* Set current transfer mode (chained mode bit added) */
      os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMA[channel].MODE, mode);
      /* Set current transfer size to 0 */
      os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMA[channel].SIZE, 0);
      /* Set the descriptor pointer */
      os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMA[channel].DPR,
			     pDevCtx->dma.channel[channel].descriptors.pci[0] | 0x1U);

      /* Make sure the writes have completed */
      os_read_register_le32(&pDevCtx->hw.adapter.pci9080->DMA[channel].DPR);
      
      /* Strobe the start bit */
      os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord, dmacsr | mask);
    } else {
      os_dbgprint(0, ("*** methodDoDmaPlxbased: DMA channel %ld not idle\n", (unsigned long) channel));
      res = MethodResult_GeneralFailure;
    }
  } else {
    /*
    ** Abort a chunk of a DMA transfer
    */

    mask = channel ? 0x0100 : 0x0001;
  
    /* First, disable the DMA channel */
    os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord, dmacsr & ~mask);
    
    /* Now check if DMA channel is active or not */
    val16 = os_read_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord);
    if (val16 & (mask << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaPlxbased: nothing to abort on DMA channel %ld\n", (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord, dmacsr | (mask << 2));
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(50);
      /* Did we succeed in aborting it? */
      val16 = os_read_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord);
      if (val16 & (mask << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(3, ("methodDoDmaPlxbased: aborted DMA on channel %ld\n", (unsigned long) channel));
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaPlxbased: failed to abort DMA on channel %ld!\n", (unsigned long) channel));
	res = MethodResult_GeneralFailure;
      }
    }
    
    /*
    ** Reenable the DMA channel and clear the DMA interrupt that will have
    ** occurred if we aborted.
    */
    os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord, dmacsr | mask | (mask << 3));
  }

  return res;
}

MethodResult
methodDoDmaXplrev0(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  os_boolean      start,
  os_uint32       mode,
  unsigned int    numBuffer)
{
  return MethodResult_NotSupported;
}

MethodResult
methodDoDmaXplrev1(
  DeviceContext* pDevCtx,
  unsigned int    channel,
  os_boolean      start,
  os_uint32       mode,
  unsigned int    numBuffer)
{
  XPLBRIDGEREGS* regs;
  MethodResult result = MethodResult_Success;
  XPLDMADESC* desc;
  os_uint32 dxctl, val32;
  os_uint32 descBus;

  regs = pDevCtx->hw.adapter.xpl;
		
  if (start) {
    /* Fix up last descriptor to end the chain and generate an interrupt */
    desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[numBuffer - 1];
    desc->NDL |= os_cpu_to_le32(XPLDMADESC_ENDCHAIN);
    
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (driverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    if (os_dbglevel(8)) {
      dumpDescriptorsXplbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.xpl,
			      channel);
    }
    
    descBus = pDevCtx->dma.channel[channel].descriptors.pci[0];
    val32 = os_read_register_le32(&regs->DMA[channel].CTL);
    if (val32 & (0x1U << 5)) {
      if (numBuffer > 1) {
	/* Set current transfer mode (adds in chained-mode bit) */
	os_write_register_le32(&regs->DMA[channel].CFG, mode | (0x1U << 9));
	/* Set the descriptor pointer & flags */
	os_write_register_le32(&regs->DMA[channel].NDL, descBus | XPLDMADESC_PCISPACE);
	os_write_register_le32(&regs->DMA[channel].NDH, 0);
      } else {
	desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[0];
	/* Set current transfer mode (DOESN'T add in chained-mode bit) */
	os_write_register_le32(&regs->DMA[channel].CFG, mode);
	/* Set current transfer PCI address */
	os_write_register_32(&regs->DMA[channel].PAL, desc->PAL);
	os_write_register_32(&regs->DMA[channel].PAH, desc->PAH);
	/* Set current transfer local address */
	os_write_register_32(&regs->DMA[channel].LA, desc->LA);
	/* Set current transfer size */
	os_write_register_32(&regs->DMA[channel].SIZ, desc->SIZ);
      }

      /* Strobe the start bit */
      dxctl = pDevCtx->shadow.xpl.dxctl[channel];
      os_write_register_le32(&regs->DMA[channel].CTL, dxctl | 0x1U);
    } else {
      os_dbgprint(0, ("*** methodDoDmaXplrev1: DMA not idle, channel=%ld\n",
		      (unsigned long) channel));
    }
  } else {
    os_dbgprint(2, ("methodDoDmaXplrev1: aborting, channel=%ld\n",
		    (unsigned long) channel));
    
    if (0) {
      dumpDescriptorsXplbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.xpl,
			      channel);
    }
    
    /* First, disable the DMA channel */
    dxctl = pDevCtx->shadow.xpl.dxctl[channel];
    dxctl &= ~0x4U;
    os_write_register_le32(&regs->DMA[channel].CTL, dxctl);
    /* Wait a bit for the bridge to abort the DMA */
    os_delay_busy_us(10);
    /* Now check if DMA channel is active or not */
    val32 = os_read_register_le32(&regs->DMA[channel].CTL);
    if (val32 & (0x1U << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaXplrev1: nothing to do, channel=%ld\n",
		      (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le32(&regs->DMA[channel].CTL, dxctl | 0x2U);
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(10);
      /* Did we succeed in aborting it? */
      val32 = os_read_register_le32(&regs->DMA[channel].CTL);
      if (val32 & (0x1U << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(1, ("methodDoDmaXplrev1: aborted, channel=%ld\n",
			(unsigned long) channel));
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaXplrev1: failed to abort, channel=%ld!\n",
			(unsigned long) channel));
	result = MethodResult_Timeout;
      }
    }
    
    /*
    ** Reenable the DMA channel and clear the DMA interrupt that
    ** will have occurred if we aborted.
    */
    dxctl |= 0x4U;
    os_write_register_le32(&regs->DMA[channel].CTL, dxctl);
  }
  
  return result;
}

/*
** -----------------------------------------------------------------------
** MethodClockWord
** -----------------------------------------------------------------------
*/

MethodResult
methodClockWordIcd2061(
  DeviceContext* pDevCtx,
  unsigned int   clockIndex,
  os_uint32      fref,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  static const unsigned long mclkRange[16][2] = {
    { 49990000,  51000000 },
    { 50990000,  53200000 },
    { 53190000,  58500000 },
    { 58490000,  60700000 },
    { 60690000,  64400000 },
    { 60390000,  66800000 },
    { 66790000,  73500000 },
    { 73490000,  75600000 },
    { 75590000,  80900000 },
    { 80890000,  83200000 },
    { 83190000,  91500000 },
    { 91490000, 100000000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 },
    { 99990000, 120100000 }
  };
  os_uint32  W;          /* programming word for ICD2061 */
  os_uint32  fvco = 0;
  os_uint32  PQERR;      /* Error so far */
  os_uint32  temp;
  os_uint32  act = 0;
  os_uint32  I;          /* VCO range */
  os_uint32  D;          /* Post-VCO divider */
  os_uint32  P;          /* VCO mult */
  os_uint32  BP = 0;     /* Best 'P' so far */
  os_uint32  Q;          /* VCO divide */
  os_uint32  BQ = 0;     /* Best 'Q' so far */
  os_uint32  QMin, QMax; /* Min/max allowable Q */
  os_boolean found = FALSE;

  if (freq < 400000 || freq > 100000000 || clockIndex >= 2) {
    return MethodResult_InvalidParameter;
  }

  /* First find a post-VCO divisor */
  for (D = 0; D < 8; D++) {
    fvco = freq * (1 << D);
    if (fvco > 49990000) {
      break;
    }
  }

  for (I = 0; I < 16; I++) {
    if ((fvco > mclkRange[I][0]) && (fvco < mclkRange[I][1])) {
      break;
    }
  }

  /* Brute force P and Q guessing */
  PQERR = 1000000;
  QMax = (fref / 200000) - 1;
  QMin = (fref / 5000000) + 1;
  for (Q = QMin; Q <= QMax; Q++) {
    for (P = 4; P < 131; P++) {
      os_uint32 err;

      /*
      ** The associativity of multiplication and divide operators is
      ** left to right. IF THIS FORMULA IS MODIFIED, ENSURE THAT IT
      ** CANNOT OVERFLOW AN (os_uint32).
      */
      temp = fref * P / Q / (1 << D) * 2;
      if (temp > freq) {
	err = temp - freq;
      } else {
	err = freq - temp;
      }
      if (err < PQERR) {
	BQ = Q;
	BP = P;
	PQERR = err;
	act = temp;
	found = TRUE;
      }
    }
  }

  if (!found) {
    return MethodResult_GeneralFailure;
  }

  /* Assemble the completed word apart from the register selection */
  W = (I << 17) | ((BP - 3) << 10) | (D << 7) | (BQ - 2);
	
  /*
  ** Add the ICD2061 register index
  **
  ** index = 0 means MCLK register
  ** index = 1 means VCLK1 register
  */
  switch (clockIndex) {
  case 0:
    W |= (3 << 21);  /* ICD2061's MCLK */
    break;
  case 1:	
    W |= (0 << 21);  /* ICD2061's VCLK1 */
    break;
  default:
    break;
  }

  *actual = act;
  *clockWord = W;

  return MethodResult_Success;
}

/*
** RDW' = RDW + 2;
** VDW' = VDW + 8;
**
** When changing any arithmetic expressions in this function, be careful
** not to allow arithmetic overflow of an unsigned 32-bit quantity.
*/
MethodResult
methodClockWordIcs307(
  DeviceContext* pDevCtx,
  unsigned int   index,
  os_uint32      fref,
  os_uint32      freq,
  os_uint32*     actual,
  os_uint32*     clockWord)
{
  const os_uint32 TTL = 0x1U;
  const os_uint32 F = 0x3U;
  const os_uint32 C = 0x0U;
  os_uint32       W; /* programming word for ICS307 */
  os_uint32       max_RDWd;
  os_uint32       min_VDWd;
  os_uint32       max_VDWd;
  os_uint32       ODd;
  os_uint32       RDWd;
  os_uint32       VDWd;
  os_uint32       best = 0;
  os_uint32       best_error;
  os_uint32       best_RDW = 0;
  os_uint32       best_VDW = 0;
  os_uint32       best_ODd = 0;
  os_uint32       best_OD = 0;
  
  if (freq < 5999000 || freq > 100001000 || index != 0) {
    return MethodResult_InvalidParameter;
  }

  /*
  ** We're using CLK2 as the local bus clock, which is
  ** CLK1/2. So we have to multiply the desired local bus
  ** clock by 2 before programming the CLK1 frequency into
  ** the ICS307
  */
  freq = freq << 1;
  
  /*
  ** Get the reference divider word.
  */
  max_RDWd = fref / 200000;
  if (max_RDWd > 129) {
    max_RDWd = 129;
  }
  
  /*
  ** Brute force guessing of best OD, RDW and VDW combination!
  */
  best_error = 0xFFFFFFFFU;
  for (ODd = 2; ODd <= 10; ODd++) {
    if (ODd == 9) {
      continue;
    }
    
    for (RDWd = 3; RDWd <= max_RDWd; RDWd++) {
      max_VDWd = 20000000 * RDWd / (fref / 10);
      if (max_VDWd > 519) {
	max_VDWd = 519;
      }
      min_VDWd = 27500000 * RDWd / fref;
      if (min_VDWd < 12) {
	min_VDWd = 12;
      }
      
      for (VDWd = min_VDWd; VDWd <= max_VDWd; VDWd++) {
	os_uint32 f, error;
	
	f = fref * 2 / RDWd * VDWd / ODd;
	error = (f > freq) ? (f - freq) : (freq - f);
	if (error < best_error) {
	  best = f;
	  best_error = error;
	  best_RDW = RDWd - 2;
	  best_VDW = VDWd - 8;
	  best_ODd = ODd;
	}
      }
    }
  }
  
  W = 0;
  W |= (best_RDW & 0x7fU)  << 0;
  W |= (best_VDW & 0x1ffU) << 7;
  switch (best_ODd) {
  case 10:
    best_OD = 0;
    break;

  case 2:
    best_OD = 1;
    break;

  case 8:
    best_OD = 2;
    break;

  case 4:
    best_OD = 3;
    break;

  case 5:
    best_OD = 4;
    break;

  case 7:
    best_OD = 5;
    break;

  case 3:
    best_OD = 6;
    break;

  case 6:
    best_OD = 7;
    break;
  }
  W |= (best_OD & 0x1ffU) << 16;
  W |= (F & 0x3U) << 19;
  W |= (TTL & 0x1U) << 21;
  W |= (C & 0x3U) << 22;
  
  *actual = best >> 1;
  *clockWord = W;
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodClockprog
** -----------------------------------------------------------------------
*/

MethodResult
methodClockProgPlxbased(
  DeviceContext* pDevCtx,
  os_uint32 freq,
  os_uint32 clockWord,
  os_boolean bBefore)
{
  os_dbgprint(3, ("methodClockProgPlxbased: entered, bBefore = %d\n", (int) bBefore));

  if (bBefore) {
    int i;
    os_uint8 d;

    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 3);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
  
    for (i = 0; i < 6; i++) {
      os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 2);
      os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
      os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 3);
      os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    }
  
    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 2);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 0);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 1);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 0);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 1);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    
    /* now send the data bits */
    for (i = 0; i < 24; i++) {
      if (clockWord & 0x1U) {
	d = 0;
      } else {
	d = 2;
      }
      clockWord >>= 1;
      
      os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, (os_uint8) (d | 0x1U));
      os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
      os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, (os_uint8) d);
      os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
      os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, (os_uint8) (d ^ 0x2U));
      os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
      os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, (os_uint8) (d ^ 0x2U) | 0x1U);
      os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    }

    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 3);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 2);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 3);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
    os_write_register_8(&pDevCtx->hw.space1.plxbased->CCON, 0);
    os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON);
  } else {
    if (os_read_register_8(&pDevCtx->hw.space1.plxbased->CCON) & 0x10) {
      return MethodResult_GeneralFailure;
    }
  }

  return MethodResult_Success;
}

MethodResult
methodClockProgXplrev0(
  DeviceContext* pDevCtx,
  os_uint32      freq,
  os_uint32      clockWord,
  os_boolean     bBefore)
{
  XPLSPACE1REGS* regs;
  os_uint32      val32;

  os_dbgprint(3, ("methodClockProgXplrev0: entered bBefore = %d\n", (int) bBefore));

  regs = pDevCtx->hw.space1.xpl;

  if (bBefore) {
    os_write_register_le32(&regs->CLOCK, (clockWord << 8) | 0x1U);
  } else {
    val32 = os_read_register_le32(&regs->CLOCK);
    if (!(val32 & 0x1)) {
      return MethodResult_Timeout;
    }
  }

  return MethodResult_Success;
}

MethodResult
methodClockProgXplrev2(
  DeviceContext* pDevCtx,
  os_uint32      freq,
  os_uint32      clockWord,
  os_boolean     bBefore)
{
  XPLSPACE1REGS* regs;
  os_uint32      newclkdiv, clockreg;
  os_uint32      val32;

  os_dbgprint(3, ("methodClockProgXplrev2: entered\n"));

  regs = pDevCtx->hw.space1.xpl;

  if (freq < 50001000) {
    /* LCLK divide of 1 for 0-50MHz */
    newclkdiv = 2;
  } else if (freq < 75001000) {
    /* LCLK divide of 1.5 for 50-75MHz */
    newclkdiv = 1;
  } else {
    /* LCLK divide of 2 for 75+MHz */
    newclkdiv = 0;
  }

  if (bBefore) {
    os_uint32 oldclkdiv;

    oldclkdiv = (os_read_register_le32(&regs->CLOCK) >> 2) & 0x3U;
    if (oldclkdiv > newclkdiv) {
      /* Increasing the frequency; change division factor at the same time as changing frequency */
      clockreg = (clockWord << 8) | (newclkdiv << 2);
    } else {
      /* Decreasing the frequency; change division factor AFTER changing frequency */
      clockreg = (clockWord << 8) | (oldclkdiv << 2);
    }
  
    os_write_register_le32(&regs->CLOCK, clockreg | 0x1U);
    os_read_register_le32(&regs->CLOCK);
  } else {
    val32 = os_read_register_le32(&regs->CLOCK);
    if (!(val32 & 0x1)) {
      return MethodResult_Timeout;
    }
  
    /* Decreasing the frequency; change division factor AFTER changing frequency */
    clockreg = (clockWord << 8) | (newclkdiv << 2);
    os_write_register_le32(&regs->CLOCK, clockreg);
    os_read_register_le32(&regs->CLOCK);
  }

  return MethodResult_Success;
}
