#if !defined(_ADATA_PLX_H)
#define _ADATA_PLX_H

/*
** EEPROM instructions: these consist of
**
**    { start_bit, opcode, address }
*/
#define EE_READ				(0x6)    /* { start_bit, opcode } for READ instruction  */
#define EE_WRITE			(0x5)    /* { start_bit, opcode } for WRITE instruction */
#define EE_WREN				(0x13)   /* { start_bit, opcode, address[n-1:n-2] } for WREN instruction */
#define EE_WDS				(0x10)   /* { start_bit, opcode, address[n-1:n-2] } for WDS instruction */
#if 0
#define EE_WRALL			(0x13)   /* 01 0001 0000 write all registers */
#define EE_PRREAD			(0x0180) /* 01 1000 0000 reads address stored in */
									 /* Protect Register */
#define EE_PRWRITE			(0x0140) /* 01 0100 0000 write address into PR */
#define EE_PREN				(0x0130) /* 01 0011 0000 protect enable instruction */
#define EE_PRCLEAR			(0x01FF) /* 01 1111 1111 clear protect register instr */
#define EE_PRDS				(0x0100) /* 01 0000 0000 ONE TIME ONLY, permenant change */
#endif

#define PLX_CNTRL_GPO		(1<<16)
#define PLX_CNTRL_GPI		(1<<17)
#define PLX_CNTRL_RST		(1<<30)

#define PLX_INTCSR_PCI_IE       (1<<8)
#define PLX_INTCSR_DB_IE        (1<<9)
#define PLX_INTCSR_PCI_ABT_IE   (1<<10)
#define PLX_INTCSR_LOC_IE       (1<<11)
#define PLX_INTCSR_PCI_ABT_INT  (1<<14)
#define PLX_INTCSR_LOC_INT      (1<<15)
#define PLX_INTCSR_DMA0_IE      (1<<18)
#define PLX_INTCSR_DMA1_IE      (1<<19)
#define PLX_INTCSR_DMA0_INT     (1<<21)
#define PLX_INTCSR_DMA1_INT     (1<<22)
#define PLX_INTCSR_BIST_INT     (1<<23)
#define PLX_INTCSR_DM_ABT_INT   (1<<24)
#define PLX_INTCSR_DMA0_ABT_INT (1<<25)
#define PLX_INTCSR_DMA1_ABT_INT (1<<26)
#define PLX_INTCSR_TGT_ABT_INT  (1<<27)

#define PLX_DMACSR_ENABLE	(1<<0)
#define PLX_DMACSR_START	(1<<1)
#define PLX_DMACSR_ABORT	(1<<2)
#define PLX_DMACSR_CLRINT	(1<<3)
#define PLX_DMACSR_DONE		(1<<4)

#define CPLD_FCON_PROG		(1<<0)
#define CPLD_FCON_INIT		(1<<1)
#define CPLD_FCON_DONE		(1<<2)

typedef struct _CONTROLXRC {
  os_uint8      FCON;
  os_uint8      CCON;
  os_uint8      IMSET;
  os_uint8      IMCLR;
  os_uint8      ICON;
  os_uint8      PSTAT;
  union {
    struct {
      /* ADM-XRCII only */
      os_uint8  MODE;
      os_uint8  FLPAGE;
    } os_declare_struct_packed admxrc2;
    struct {
      os_uint8  reserved1[2];
    } os_declare_struct_packed other;
  } os_declare_struct_packed variant;
  os_uint8      SELECTMAP[8];
} os_declare_struct_packed CONTROLXRC;

typedef struct _CONTROLWRC2 {
  os_uint8      FCON;
  os_uint8      CCON;
  os_uint8      IMSET;
  os_uint8      IMCLR;
  os_uint8      ICON;
  os_uint8      PSTAT;
  os_uint8      MODE;
  os_uint8      reserved1[1];
  os_uint8      SELECTMAP[4];
  os_uint32     FLADDR;
  os_uint8      FLDATA;
  os_uint8      reserved2[3];
} os_declare_struct_packed CONTROLWRC2;

typedef struct _DMACON9080 {
  os_uint32 MODE;
  os_uint32 PADR;
  os_uint32 LADR;
  os_uint32 SIZE;
  os_uint32 DPR;
} os_declare_struct_packed DMACON9080;

typedef struct _PLX9080REGS {
  os_uint32   LAS0RR;       /*  00 */
  os_uint32   LAS0BA;       /*  04 */
  os_uint32   MARBR;        /*  08 */
  os_uint32   BIGEND;       /*  0C */
  os_uint32   EROMRR;       /*  10 */
  os_uint32   EROMBA;       /*  14 */
  os_uint32   LBRD0;        /*  18 */
  os_uint32   DMRR;         /*  1C */
  os_uint32   DMLBAM;       /*  20 */
  os_uint32   DMLBAI;       /*  24 */
  os_uint32   DMPBAM;       /*  28 */
  os_uint32   DMCFGA;       /*  2C */
  os_uint32   OPLFIS;       /*  30 */
  os_uint32   OPLFIM;       /*  34 */
  os_uint32   reserved1[2]; /*  38 */
  os_uint32   MAILBOX[8];   /*  40-5F */
  os_uint32   P2LDBELL;     /*  60 */
  os_uint32   L2PDBELL;     /*  64 */
  os_uint32   INTCSR;       /*  68 */
  os_uint32   CNTRL;        /*  6C */
  os_uint32   PCIHDR;       /*  70 */
  os_uint16   PCIHREV;      /*  74 */
  os_uint16   RES76;        /*  76 */
  os_uint32   MBOX0;        /*  78  always present */
  os_uint32   MBOX1;        /*  7C  always present */
  DMACON9080  DMA[2];       /*  80 */
  union {
    os_uint8  asByte[2];    /*  A8,A9 */
    os_uint16 asWord;       /*  A8,A9 */
  } os_declare_struct_packed DMACSR;
  os_uint8    reserved2[2]; /*  AA,AB */
  os_uint32   DMAARB;       /*  AC */
  os_uint32   DMATHR;       /*  B0 */
  os_uint32   reserved3[3]; /*  B4-BF */
  os_uint32   MQCR;         /*  C0 */
  os_uint32   QBAR;         /*  C4 */
  os_uint32   IFHPR;        /*  C8 */
  os_uint32   IFTPR;        /*  CC */
  os_uint32   IPHPR;        /*  D0 */
  os_uint32   IPTPR;        /*  D4 */
  os_uint32   OFHPR;        /*  D8 */
  os_uint32   OFTPR;        /*  DC */
  os_uint32   OPHPR;        /*  E0 */
  os_uint32   OPTPR;        /*  E4 */
  os_uint32   QSR;          /*  E8 */
  os_uint32   reserved4;    /*  EC */
  os_uint32   LAS1RR;       /*  F0 */
  os_uint32   LAS1BA;       /*  F4 */
  os_uint32   LBRD1;        /*  F8 */
  os_uint32   reserved5;    /*  FC */
} os_declare_struct_packed PLX9080REGS;

/*
** The order of members in this structure assumes that
** the valid bit in the MODE member will always be 0.
*/
typedef struct DMACON9656 {
  os_uint32   MODE;
  os_uint32   PADR;
  os_uint32   LADR;
  os_uint32   SIZE;
  os_uint32   DPR;
} os_declare_struct_packed DMACON9656;

typedef struct _PLX9656REGS {
  os_uint32	LAS0RR;			/*  00 */
  os_uint32	LAS0BA;			/*  04 */
  os_uint32	MARBR;			/*  08 */
  os_uint8	BIGEND;			/*  0C */
  os_uint8	LMISC1;			/*  0D */
  os_uint8	PROT_AREA;		/*  0E */
  os_uint8	LMISC2;			/*  0F */
  os_uint32	EROMRR;			/*  10 */
  os_uint32	EROMBA;			/*  14 */
  os_uint32	LBRD0;			/*  18 */
  os_uint32	DMRR;			/*  1C */
  os_uint32	DMLBAM;			/*  20 */
  os_uint32	DMLBAI;			/*  24 */
  os_uint32	DMPBAM;			/*  28 */
  os_uint32	DMCFGA;			/*  2C */
  os_uint32	OPQIS;			/*  30 */
  os_uint32	OPQIM;			/*  34 */
  os_uint8	reserved1[8];           /*  38 */
  union {
    struct _struct_ni20_A {
      os_uint32	MBOX0;	/*  40 */
      os_uint32	MBOX1;	/*  44 */
      os_uint32	MBOX2;	/*  48 */
      os_uint32	MBOX3;	/*  4C */
      os_uint32	MBOX4;	/*  50 */
      os_uint32	MBOX5;	/*  54 */
      os_uint32	MBOX6;	/*  58 */
      os_uint32	MBOX7;	/*  5C */
    } os_declare_struct_packed NoI2O;
    struct _struct_wi20_A {
      os_uint32	IQP;	/*  40 */
      os_uint32	OQP;	/*  44 */
      os_uint32	MBOX2;	/*  48 */
      os_uint32	MBOX3;	/*  4C */
      os_uint32	MBOX4;	/*  50 */
      os_uint32	MBOX5;	/*  54 */
      os_uint32	MBOX6;	/*  58 */
      os_uint32	MBOX7;	/*  5C */
    } os_declare_struct_packed WithI2O;
  } os_declare_struct_packed MBOXA;
  os_uint32	P2LDBELL;       /*  60 */
  os_uint32	LD2PBELL;       /*  64 */
  os_uint32	INTCSR;         /*  68 */
  os_uint32	CNTRL;          /*  6C */
  os_uint32	PCIHIDR;        /*  70 */
  os_uint8	PCIHREV;        /*  74 */
  os_uint8	reserved2[3];   /*  75 */
  union {
    struct _struct_ni20_B {
      os_uint8 reserved3[8];
    } os_declare_struct_packed NoI2O;
    struct _struct_wi20_B {
      os_uint32	MBOX0;          /*  78 */
      os_uint32	MBOX1;          /*  7C */
    } os_declare_struct_packed WithI2O;
  } os_declare_struct_packed MBOXB;
  DMACON9656    DMA[2];		/*  80 */
  os_uint8      DMACSR[2];	/*  A8 */
  os_uint8	reserved4[2];	/*  AA */
  os_uint32	DMAARB;		/*  AC */
  os_uint32	DMATHR;		/*  B0 */
  os_uint32	DMADAC0;	/*  B4 */
  os_uint32	DMADAC1;	/*  B8 */
  os_uint8	reserved5[4];	/*  BC */
  os_uint32	MQCR;		/*  C0 */
  os_uint32	QBAR;		/*  C4 */
  os_uint32	IFHPR;		/*  C8 */
  os_uint32	IFTPR;		/*  CC */
  os_uint32	IPHPR;		/*  D0 */
  os_uint32	IPTPR;		/*  D4 */
  os_uint32	OFHPR;		/*  D8 */
  os_uint32	OFTPR;		/*  DC */
  os_uint32	OPHPR;		/*  E0 */
  os_uint32	OPTPR;		/*  E4 */
  os_uint32	QSR;		/*  E8 */
  os_uint8	reserved6[4];	/*  EC */
  os_uint32	LAS1RR;		/*  F0 */
  os_uint32	LAS1BA;		/*  F4 */
  os_uint32	LBRD1;		/*  F8 */
  os_uint32	DMDAC;		/*  FC */
  os_uint32	PCIARB;		/* 100 */
  os_uint32	PABTADR;	/* 104 */
  os_uint32     reserved7[62];  /* 108-1FF */
} os_declare_struct_packed PLX9656REGS;

/*
** PCI9080 DMA DPRx bit definitions                    
*/
#define PLXDESCF_LOCAL       (0 << 0)
#define PLXDESCF_PCI         (1 << 0)
#define PLXDESCF_ENDCHAIN    (1 << 1)
#define PLXDESCF_INTERRUPT   (1 << 2)
#define PLXDESCF_PCI2LOC     (0 << 3)
#define PLXDESCF_LOC2PCI     (1 << 3)

typedef struct _PLXDESC {
  os_uint32 pci;
  os_uint32 local;
  os_uint32 count;
  os_uint32 link;
} os_declare_struct_packed PLXDESC;

#endif
