/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** spaceconfig.c - IOCTL handlers for getting/setting local bus space
**                 configurations.
**
** ----------------------------------------------------------------------------
*/

#include "device.h"
#include "ioctl.h"
#include "methods.h"
#include "spaceconfig.h"

OS_STATUS
ADMXRC_GetSpaceConfig(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_GETSPACECONFIG* in;
  ADMXRC_IOCTLS_GETSPACECONFIG* out;
  unsigned int                  space;
  os_uint32                     flags;
  MethodResult                  result;
  
  in = (ADMXRC_IOCTLS_GETSPACECONFIG*) req->param.inBuf;
  if (in == NULL || req->param.inSize != sizeof(in->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }
  out = (ADMXRC_IOCTLS_GETSPACECONFIG*) req->param.outBuf;
  if (out == NULL || req->param.outSize != sizeof(out->out)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  space = in->in.space_index;

  /* Read the configuration */
  result = pDevCtx->methods.configSpace(pDevCtx, FALSE, space, &flags);
  if (result == MethodResult_Success) {
    out->out.flags = flags;
    req->status.length = sizeof(out->out);
    return OS_STATUS_SUCCESS;
  } else if (result == MethodResult_InvalidParameter) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  } else if (result == MethodResult_NotSupported) {
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  } else {
    return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
  }
}

OS_STATUS
ADMXRC_SetSpaceConfig(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_SETSPACECONFIG* in;
  unsigned int                  space;
  os_uint32                     flags;
  MethodResult                  result;

  in = (ADMXRC_IOCTLS_SETSPACECONFIG*) req->param.inBuf;
  if (in == NULL || req->param.inSize != sizeof(in->in)) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  }

  space = in->in.space_index;
  flags = in->in.flags;

  /* Write the configuration into registers */
  result = pDevCtx->methods.configSpace(pDevCtx, TRUE, space, &flags);
  if (result == MethodResult_Success) {
    return OS_STATUS_SUCCESS;
  } else if (result == MethodResult_InvalidParameter) {
    return OS_MAKE_STATUS(ADMXRC_INVALID_PARAMETER);
  } else if (result == MethodResult_NotSupported) {
    return OS_MAKE_STATUS(ADMXRC_NOT_SUPPORTED);
  } else {
    return OS_MAKE_STATUS(ADMXRC_UNKNOWN_ERROR);
  }
}
