#ifndef _ADATA_ADMXRC_H
#define _ADATA_ADMXRC_H

/*
** admxrc.h -  ADM-XRC SDK API header file (ADMXRC interface)
**
** (C) Copyright 1999-2002 Alpha Data Parallel Systems Ltd.
*/

#include <stdint.h>
#include <admxrc/status.h>
#include <admxrc/struct.h>

/*
** Handle to a card
*/
typedef void *ADMXRC_HANDLE;

/*
** Error handler function type
*/
typedef void (*ADMXRC_HANDLER_FUNCTION)(const char* FnName, ADMXRC_STATUS Status);

/*
** FPGA Image file
*/
typedef void* ADMXRC_IMAGE;

/*
** Function prototypes
*/
#ifdef __cplusplus
extern "C"{
#endif

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_CloseCard(
    ADMXRC_HANDLE                        card);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_OpenCard(
    ADMXRC_DEVICE_NUM                    cardID,
    ADMXRC_HANDLE*                       card);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_GetVersionInfo(
    ADMXRC_HANDLE                        card,
    ADMXRC_VERSION_INFO*                 info);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_GetCardInfo(
    ADMXRC_HANDLE                        card,
    ADMXRC_CARD_INFO*                    info);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_SetClockRate(
    ADMXRC_HANDLE                        card,
    ADMXRC_CLOCK                         clock,
    double                               rate);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_ConfigureFromBuffer(
    ADMXRC_HANDLE                        card,
    void*                                buffer,
    uint32_t                             length);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_ConfigureFromBufferDMA(
    ADMXRC_HANDLE                        card,
    void*                                buffer,
    uint32_t                             length,
    uint32_t                             dma_channel,
    void**                               ignored);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_ConfigureFromFile(
    ADMXRC_HANDLE                        card,
    char*                                fileName);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_ConfigureFromFileDMA(
    ADMXRC_HANDLE                        card,
    char*                                fileName,
    uint32_t                             dma_channel,
    void**                               event);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_InstallErrorHandler(
    ADMXRC_HANDLER_FUNCTION              handler);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_ProgramBytes(
    ADMXRC_HANDLE                        card,
    unsigned long*                       program,
    unsigned long                        length);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_WriteReg(
    ADMXRC_HANDLE                        card,
    unsigned char                        index,
    unsigned char                        value);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_ReadReg(
    ADMXRC_HANDLE                        card,
    unsigned char                        index,
    unsigned char*                       value);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_GetClockType(
    ADMXRC_HANDLE                        card,
    ADMXRC_CLOCK_TYPE*                   value);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_GetBaseAddress(
    ADMXRC_HANDLE                        card,
    void**                               pFPGA);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_WaitForInterrupt(
    ADMXRC_HANDLE                        card,
    unsigned long*                       timeout_us,
    unsigned long                        flags,
    void**                               ignored);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_Read(
    ADMXRC_HANDLE                        card,
    unsigned long                        width,
    unsigned long                        flags,
    uint32_t                                local,
    void*                                data,
    unsigned long                        len);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_Write(
    ADMXRC_HANDLE                        card,
    unsigned long                        width,
    unsigned long                        flags,
    uint32_t                             local,
    void*                                data,
    unsigned long                        len);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_SetupDMA(
    ADMXRC_HANDLE                        card,
    void*                                buffer,
    unsigned long                        size,
    uint32_t                             flags,
    ADMXRC_DMADESC*                      dmadesc);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_UnsetupDMA(
    ADMXRC_HANDLE                        card,
    ADMXRC_DMADESC                       dmadesc);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_MapDirectMaster(
    ADMXRC_HANDLE                        card,
    ADMXRC_DMADESC                       dmadesc,
    unsigned long                        offset,
    unsigned long                        length,
    ADMXRC_BUFFERMAP*                    map);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_SyncDirectMaster(
    ADMXRC_HANDLE                        card,
    ADMXRC_DMADESC                       dmadesc,
    unsigned long                        offset,
    unsigned long                        length,
    uint32_t                             syncmode);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_DoDMA(
    ADMXRC_HANDLE                        card,
    ADMXRC_DMADESC                       dmadesc,
    unsigned long                        offset,
    unsigned long                        length,
    uint32_t                             local,
    uint32_t                             direction,
    uint32_t                             channel,
    uint32_t                             mode,
    uint32_t                             flags,
    uint32_t*                            timeout,
    void**                               ignored);
  
ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_DoDMAImmediate(
    ADMXRC_HANDLE                        card,
    void*                                buffer,
    unsigned long                        length,
    uint32_t                             local,
    uint32_t                             direction,
    uint32_t                             channel,
    uint32_t                             mode,
    uint32_t                             flags,
    uint32_t*                            timeout,
    void**                               ignored);

/*
** Utility functions
*/

ADMXRC_DLLFUNC uint32_t
ADMXRC_BuildDMAModeWord(
    uint32_t                             buswidth,
    uint32_t                             waitstates,
    uint32_t                             misc);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_StatusToString(
    ADMXRC_STATUS                        status,
    char*                                string,
    unsigned long                        length);

ADMXRC_DLLFUNC const char*
ADMXRC_GetStatusString(
    ADMXRC_STATUS                        status);

ADMXRC_DLLFUNC void
ADMXRC_ReverseBytes(
    ADMXRC_IMAGE                         image,
    unsigned long                        offset,
    unsigned long                        length);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_FindImageOffset(
    ADMXRC_FPGA_TYPE                     fpgatype,
    ADMXRC_IMAGE                         image,
    unsigned long                        size,
    unsigned long*                       offs);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_LoadFpgaFile(
    char*                                name,
    ADMXRC_IMAGE*                        image,
    unsigned long*                       bytesread);

ADMXRC_DLLFUNC ADMXRC_STATUS
ADMXRC_UnloadFpgaFile(
    ADMXRC_IMAGE                         image);

#ifdef __cplusplus
} 
#endif

#endif /* ADMXRC_H */

