/*
** struct.h - structures used in ADM-XRC API calls
**
** (C) Copyright 2000 Alpha Data Parallel Systems Ltd.
*/

#ifndef _ADATA_ADMXRC_STRUCT_H
#define _ADATA_ADMXRC_STRUCT_H

/*
** A value to indicate which card
*/
typedef unsigned long ADMXRC_DEVICE_NUM;

/*
** Handle to a DMA descriptor
*/
typedef unsigned long ADMXRC_DMADESC;

/*
** DMA direction options
*/
typedef enum
{
  ADMXRC_PCI2LOCAL = 0,
  ADMXRC_LOCAL2PCI = 1,
  ADMXRC_DMA_DIRECTION_FORCE32BITS = 0x7FFFFFFF
} ADMXRC_DMA_DIRECTION;

/*
** DMA width flags for ADMXRC_BuildDmaModeWord()
*/
typedef enum _ADMXRC_DMA_WIDTH
{
  ADMXRC_DMAWIDTH_8  = 0,
  ADMXRC_DMAWIDTH_16 = 1,
  ADMXRC_DMAWIDTH_32 = 2,
  ADMXRC_DMA_WIDTH_FORCE32BITS = 0x7FFFFFFF
} ADMXRC_DMA_WIDTH;

/*
** Misc DMA mode options for ADMXRC_BuildDmaModeWord()
*/
#define ADMXRC_DMAMODE_IGNOREREADY		(0x0 << 6)	/* default */
#define ADMXRC_DMAMODE_USEREADY			(0x1 << 6)
#define ADMXRC_DMAMODE_IGNOREBTERM		(0x0 << 7)	/* default */
#define ADMXRC_DMAMODE_USEBTERM			(0x1 << 7)
#define ADMXRC_DMAMODE_BURSTDISABLE		(0x0 << 8)	/* default */
#define ADMXRC_DMAMODE_BURSTENABLE		(0x1 << 8)
#define ADMXRC_DMAMODE_INCREMENTLOCAL           (0x0 << 11)	/* default */
#define ADMXRC_DMAMODE_FIXEDLOCAL		(0x1 << 11)
#define ADMXRC_DMAMODE_NONDEMAND		(0x0 << 12)	/* default */
#define ADMXRC_DMAMODE_DEMAND			(0x1 << 12)

/*
** DMA channel options
*/
typedef enum _ADMXRC_DMA_CHANNEL
{
  ADMXRC_DMACHAN_0 = 0,
  ADMXRC_DMACHAN_1 = 1,
  ADMXRC_DMACHAN_ANY = 0xFFU,
  ADMXRC_DMA_CHANNEL_FORCE32BITS = 0x7FFFFFFF
} ADMXRC_DMA_CHANNEL;

/*
** ADMXRC_DoDMA() flags
*/
#define ADMXRC_DMAFLAG_CANQUEUE			(0x0 << 0)	/* default */
#define ADMXRC_DMAFLAG_DONOTQUEUE		(0x1 << 0)

/*
** Clock selections
*/
typedef enum _ADMXRC_CLOCK
{
  ADMXRC_MCLK,
  ADMXRC_VCLK,
  ADMXRC_PCICLK,
  ADMXRC_VCLK1,
  ADMXRC_VCLK2,
  ADMXRC_VCLK3,
  ADMXRC_CLOCK_FORCE32BITS = 0x7FFFFFFF
} ADMXRC_CLOCK;

/*
 * ADMXRC_Read() and ADMXRC_Write() flags
 */
#define ADMXRC_IOBYTE	(0)
#define ADMXRC_IOWORD	(1)
#define ADMXRC_IOLONG	(2)

#define ADMXRC_IOLOCAL	(0<<0)
#define ADMXRC_IOFIXED	(1<<0)
#define ADMXRC_IOPLX    (1<<1)

/*
** FPGA Type
*/
typedef enum _ADMXRC_FPGA_TYPE
{
  ADMXRC_FPGA_4085XL,
  ADMXRC_FPGA_40150XV,
  ADMXRC_FPGA_40200XV,
  ADMXRC_FPGA_40250XV,
  ADMXRC_FPGA_V1000,
  ADMXRC_FPGA_V400,
  ADMXRC_FPGA_V600,
  ADMXRC_FPGA_V800,
  ADMXRC_FPGA_V2000E,
  ADMXRC_FPGA_V1000E,
  ADMXRC_FPGA_V1600E,
  ADMXRC_FPGA_V3200E,
  ADMXRC_FPGA_V812E,
  ADMXRC_FPGA_V405E,
  ADMXRC_FPGA_UNKNOWN,
  ADMXRC_FPGA_TYPE_FORCE32BITS = 0x7FFFFFFF
} ADMXRC_FPGA_TYPE;

/*
** Clock oscillator types
*/
typedef enum _ADMXRC_CLOCK_TYPE
{
  ADMXRC_CLOCKTYPE_16		= 0, /* 16.666MHz */
  ADMXRC_CLOCKTYPE_14		= 1, /* 14.318MHz */
  ADMXRC_CLOCK_TYPE_FORCE32BITS = 0x7FFFFFFF
} ADMXRC_CLOCK_TYPE;

/*
** Card information structure
*/
typedef struct _ADMXRC_CARD_INFO
{
  ADMXRC_DEVICE_NUM CardID;
  unsigned long     RAMBankFitted[4];
  ADMXRC_FPGA_TYPE  FPGAType;
  unsigned long     PhysicalMemoryBase;
  unsigned long*    MemoryBase;
  unsigned long     BoardRevision;
  unsigned long     LogicRevision;
  unsigned long     SerialNum;
  unsigned long     Timeout;
} ADMXRC_CARD_INFO;

/*
** Structure returning version information
*/
typedef struct _ADMXRC_VERSION_INFO
{
  unsigned char DriverMinor;
  unsigned char DriverMajor;
  unsigned char APIMinor;
  unsigned char APIMajor;
} ADMXRC_VERSION_INFO;

/*
** A structure used to map a user buffer for Direct Master
** access by the FPGA.
*/
typedef struct _ADMXRC_BUFFERMAP
{
  unsigned long  MaxPages;     /* User initialises to length of PCI address buffer      */
  unsigned long* PagesPci;     /* User initialises to point to PCI address buffer       */
  unsigned long  PageLength;   /* Receives size in bytes of a page                      */
  unsigned long  PageBits;     /* Number of bits representing a page offset             */
  unsigned long  PagesSpanned; /* Receives number of pages spanned                      */
  unsigned long  BytesSpanned; /* Receives number of bytes spanned                      */
  unsigned long  InitOffset;   /* Receives page offset within 1st page of start of span */
} ADMXRC_BUFFERMAP;

/*
** Type of syncing to be done by ADMXRC_SyncDirectMaster()
*/
typedef enum _ADMXRC_SYNCMODE
{
  ADMXRC_SYNC_CPUTOFPGA = 0x1, /* CPU has set up buffer for FPGA to access */
  ADMXRC_SYNC_FPGATOCPU = 0x2, /* FPGA has modified buffer                 */
  ADMXRC_SYNCMODE_FORCE32BITS = 0x7FFFFFFF
} ADMXRC_SYNCMODE;

#endif

