/*
** lib_generic.c
**
** Platform Independent API Library Calls 
**
** (C) Copyright 2002-2004 Alpha Data Parallel Systems Ltd.
*/

#include "platform2.h"
#include <admxrc2.h>

#include "lib_ver.h"

typedef struct _ADMXRC2_STATUS_MSG
{
  ADMXRC2_STATUS	code;
  char*			description;
} ADMXRC2_STATUS_MSG;

#define ADMXRC2_LASTERROR (-1)

static const ADMXRC2_STATUS_MSG statusMessage[] =
{
  { ADMXRC2_SUCCESS,           "success" },
  { ADMXRC2_INTERNAL_ERROR,    "an error in the API logic occurred" },
  { ADMXRC2_NO_MEMORY,         "failed to allocate memory" },
  { ADMXRC2_CARD_NOT_FOUND,    "failed to open specified card" },
  { ADMXRC2_FILE_NOT_FOUND,    "failed to open file" },
  { ADMXRC2_INVALID_FILE,      "file appears to be corrupt" },
  { ADMXRC2_FPGA_MISMATCH,     "the bitstream target FPGA did not match the FPGA on the card" },
  { ADMXRC2_INVALID_HANDLE,    "the card handle passed was invalid" },
  { ADMXRC2_TIMEOUT,           "the operation was not completed with the timeout period" },
  { ADMXRC2_CARD_BUSY,         "the card could not be opened because it was in use" },
  { ADMXRC2_INVALID_PARAMETER, "an invalid parameter was passed" },
  { ADMXRC2_CLOSED,            "the card handle was closed before the operation was completed" },
  { ADMXRC2_CARD_ERROR,        "the card could not be opened because the wrong card ID was specified" },
  { ADMXRC2_NOT_SUPPORTED,     "the requested operation is not supported" },
  { ADMXRC2_DEVICE_BUSY,       "the operation could not be performed because the device or resource was busy" },
  { ADMXRC2_INVALID_DMADESC,   "an invalid handle to a DMA descriptor was passed" },
  { ADMXRC2_NO_DMADESC,        "unable to allocate a DMA descriptor" },
  { ADMXRC2_FAILED,            "the operation failed" },
  { ADMXRC2_PENDING,           "the operation is still in progress" },
  { ADMXRC2_UNKNOWN_ERROR,     "an error occurred whose origin is unknown" },
  { ADMXRC2_NULL_POINTER,      "a NULL pointer was passed" },
  { ADMXRC2_CANCELLED,         "the operation was cancelled because the requesting thread terminated" },
  { ADMXRC2_BAD_DRIVER,        "there is a problem with the device driver" },
  { ADMXRC2_LASTERROR,         "" }
};

static const char fpgaNames[][32]  =
{
  "XC4085XL",   /*  0 */
  "XC40150XV",  /*  1 */
  "XC40200XV",  /*  2 */
  "XC40250XV",  /*  3 */
  "v1000",      /*  4 */
  "v400",       /*  5 */
  "v600",       /*  6 */
  "v800",       /*  7 */
  "v2000e",     /*  8 */
  "v1000e",     /*  9 */
  "v1600e",     /* 10 */
  "v3200e",     /* 11 */
  "v812e",      /* 12 */
  "v405e",      /* 13 */
  "unknown14",  /* 14 */
  "unknown15",  /* 15 */
  "unknown16",  /* 16 */
  "unknown17",  /* 17 */
  "unknown18",  /* 18 */
  "unknown19",  /* 19 */
  "unknown20",  /* 20 */
  "unknown21",  /* 21 */
  "unknown22",  /* 22 */
  "unknown23",  /* 23 */
  "unknown24",  /* 24 */
  "unknown25",  /* 25 */
  "unknown26",  /* 26 */
  "unknown27",  /* 27 */
  "unknown28",  /* 28 */
  "unknown29",  /* 29 */
  "unknown30",  /* 30 */
  "unknown31",  /* 31 */
  "2v1000",     /* 32 */
  "2v1500",     /* 33 */
  "2v2000",     /* 34 */
  "2v3000",     /* 35 */
  "2v4000",     /* 36 */
  "2v6000",     /* 37 */
  "2v8000",     /* 38 */
  "2v10000",    /* 39 */
  "unknown40",  /* 40 */
  "unknown41",  /* 41 */
  "unknown42",  /* 42 */
  "unknown43",  /* 43 */
  "unknown44",  /* 44 */
  "unknown45",  /* 45 */
  "unknown46",  /* 46 */
  "unknown47",  /* 47 */
  "unknown48",  /* 48 */
  "unknown49",  /* 49 */
  "unknown50",  /* 50 */
  "unknown51",  /* 51 */
  "unknown52",  /* 52 */
  "unknown53",  /* 53 */
  "unknown54",  /* 54 */
  "unknown55",  /* 55 */
  "unknown56",  /* 56 */
  "unknown57",  /* 57 */
  "unknown58",  /* 58 */
  "unknown59",  /* 59 */
  "unknown60",  /* 60 */
  "unknown61",  /* 61 */
  "unknown62",  /* 62 */
  "unknown63",  /* 63 */
  "2vp2",       /* 64 */
  "2vp4",       /* 65 */
  "2vp7",       /* 66 */
  "2vp20",      /* 67 */
  "2vp30",      /* 68 */
  "2vp40",      /* 69 */
  "2vp50",      /* 70 */
  "2vp100",     /* 71 */
  "2vp125",     /* 72 */
  "2vp70"       /* 73 */
};

ADMXRC2_ERROR_HANDLER _ADMXRC2_errorHandler = NULL;

ADMXRC2_STATUS
_ADMXRC2_handleError(
  const char*    function,
  ADMXRC2_STATUS statusIn)
{
  if (_ADMXRC2_errorHandler != NULL) {
    _ADMXRC2_errorHandler(function, statusIn);
  }

  return statusIn;
}

static ADMXRC2_STATUS
Unconfigure(
  ADMXRC2_HANDLE card)
{
  ADMXRC2_STATUS status;

  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_CHECKFPGACONFIGURED,
		       NULL,
		       0,
		       NULL,
		       0,
		       &status);
  return status;
}

static ADMXRC2_STATUS
CheckConfigured(
  ADMXRC2_HANDLE card)
{
  ADMXRC2_STATUS status;

  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_CHECKFPGACONFIGURED,
		       NULL,
		       0,
		       NULL,
		       0,
		       &status);
  return status;
}

/*
** Swap the bits in each byte as Virtex selectmap
** interface bit-ordering does not match bit files.
*/
static void
ReverseBytes(
  ADMXRC2_IMAGE buf,
  unsigned long size)
{
  int	j;
  char	d, dd;
  char*	p = (char*) buf;

  while (size--) {
    d = *p;
    dd = 0;
    for (j = 0; j < 8; j++) {
      dd <<= 1;
      if (d & 1)
	dd |= 1;
      d >>= 1;
    }
    *p++ = dd;
  }
}

static ADMXRC2_STATUS
SelectMapOffset(
  ADMXRC2_CARD_INFO* info,
  FILE*              file,
  unsigned long*     image_size,
  unsigned long*     file_pos)
{
  typedef enum {
    Preamble = 0,
    Records  = 1,
    Quit     = 2
  } State;
  const int        blocksize = 1024;
  char             fileFpgaName[32];
  char             fpgaName[32];
  char             tmpbuf[1024];
  State            state;
  ADMXRC2_STATUS   status;
  unsigned long    filesize, left, pos;
  os_boolean       found;
  unsigned long    offs, level;
  os_boolean       more;
  unsigned long    skip;

  if (info->FPGAType >= ADMXRC2_FPGA_UNKNOWN) {
    status = ADMXRC2_FPGA_MISMATCH; 
    goto failed;
  }

  strcpy(fpgaName, fpgaNames[info->FPGAType]);
  switch (info->BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
    strcat(fpgaName, "bg560");
    break;
  case ADMXRC2_BOARD_ADMXRC2_LITE:
    strcat(fpgaName, "fg456");
    break;
  case ADMXRC2_BOARD_ADMXRC2:
    strcat(fpgaName, "ff1152");
    break;
  case ADMXRC2_BOARD_ADMXRC2_PRO_LITE:
    strcat(fpgaName, "ff896");
    break;
  case ADMXRC2_BOARD_ADPWRC2:
    strcat(fpgaName, "ff1517");
    break;
  case ADMXRC2_BOARD_ADPDRC2:
    strcat(fpgaName, "ff1152");
    break;
  case ADMXRC2_BOARD_ADMXRC2_PRO:
    strcat(fpgaName, "ff1704");
    break;
  default:
    status = ADMXRC2_FPGA_MISMATCH;
    goto failed;
  }

  /* Convert the FPGA names to lower case before comparing them	*/
  _ADMXRC2_strlwr(fpgaName);

  if (fseek(file, 0, SEEK_END) != 0) {
    status = ADMXRC2_INVALID_FILE;
    goto failed;
  }

  if ((filesize = ftell(file)) < 0) {
    status = ADMXRC2_INVALID_FILE;
    goto failed;
  }
  rewind(file);

  /*
  ** Find the beginning of the SelectMap data in the file
  */
  left = filesize;
  state = Preamble;
  found = FALSE;
  pos = 0;
  offs = 0;
  level = 0;
  more = TRUE;
  skip = 0;
  while (!found) {
    if (more) {
      unsigned long chunk;
      unsigned long numread;
			
      if (!left) {
	status = ADMXRC2_INVALID_FILE;
	goto failed;
      }
      memmove(tmpbuf + 0, tmpbuf + offs, level);
      offs = level;
      chunk = blocksize - level;
      if (left < chunk) {
	chunk = left;
      }
      numread = fread(tmpbuf, 1, chunk, file);
      if (!numread) {
	status = ADMXRC2_INVALID_FILE;
	goto failed;
      }
      level += chunk;
    }
    more = FALSE;
    if (skip) {
      unsigned long chunk;
			
      chunk = level;
      if (chunk > skip) {
	chunk = skip;
      }
			
      level -= chunk;
      offs += chunk;
      skip -= chunk;
      pos += chunk;
      
      if (skip) {
	more = TRUE;
      }
    } else {
      switch (state) {
      case Preamble:
	if (level < 2) {
	  more = TRUE;
	} else {
	  /* Skip over the preamble */
	  skip = ((unsigned short) tmpbuf[offs] << 8) + tmpbuf[offs + 1];
	  skip += 2;
	  offs += 2;
	  level -= 2;
	  pos += 2;
	  state = Records;
	}
	break;
      case Records:
	if (level < 1) {
	  more = TRUE;
	} else {
	  char          rectype;
	  unsigned long reclen;
					
	  rectype = tmpbuf[offs];
	  switch (rectype) {
	  case 0x62:
	    /*
	    ** Part type record
	    */
	    if (level < 3) {
	      more = TRUE;
	    } else {
	      reclen = ((unsigned short) tmpbuf[offs + 1] << 8) + tmpbuf[offs + 2];
	      if (reclen >= sizeof(fileFpgaName)) {
		status = ADMXRC2_INVALID_FILE;
		goto failed;
	      }
	      if (level < (3 + reclen)) {
		more = TRUE;
	      } else {
		strncpy(fileFpgaName, tmpbuf + offs + 3, reclen);
		fileFpgaName[reclen] = '\0';
		_ADMXRC2_strlwr(fileFpgaName);
		/*
		** Compare the FPGA names for a match
		*/
		if (strcmp(fileFpgaName, fpgaName)) {
		  status = ADMXRC2_FPGA_MISMATCH;
		  goto failed;
		}
		offs += (3 + reclen);
		level -= (3 + reclen);
		pos += (3 + reclen);
	      }
	    }
	    break;
	  case 0x65:
	    /*
	    ** End of header record
	    */
	    skip = 5;
	    state = Quit;
	    break;
	  case 0x61:
	  case 0x63:
	  case 0x64:
	    /*
	    ** Unknown record types
	    */
	    if (level < 3) {
	      more = TRUE;
	    } else {
	      reclen = ((unsigned short) tmpbuf[offs + 1] << 8) + tmpbuf[offs + 2];
	      if (level < (3 + reclen)) {
		more = TRUE;
	      } else {
		offs += (3 + reclen);
		level -= (3 + reclen);
		pos += (3 + reclen);
	      }
	    }
	    break;
	  default:
	    /*
	    ** Unexpected record type
	    */
	    status = ADMXRC2_INVALID_FILE;
	    goto failed;
	    break;
	  }
	}
	break;
      case Quit:
	found = TRUE;
	break;
      }
    }
  }

  *image_size = filesize - pos;
  *file_pos = pos;
  status = ADMXRC2_SUCCESS;

 failed:
  return status;
}



/*
** --------------------------------------------------------------------
** Exported functions start here
** --------------------------------------------------------------------
*/

/*
** Put together a partial mode word that can be passed to ADMXRC_DoDMA.
**
** Parameters are validated, and if not valid, the special invalid mode word
** 0xffffffffU is returned. Otherwise, a partial mode word for the PLX.DMAMODEn
** registers is constructed, MINUS the flags that the driver adds, and returned.
**
** The driver always adds other flags which are not available to the application,
** such as "demand mode" and "chain mode" before programming the PLX.DMAMODEn
** register.
*/
uint32_t 
ADMXRC2_BuildDMAModeWord(
  ADMXRC2_BOARD_TYPE cardtype,
  ADMXRC2_IOWIDTH    buswidth,
  unsigned int       waitstates,
  uint32_t           misc)
{
  unsigned long mode = 0;

  switch (cardtype) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
    switch (buswidth) {
    case ADMXRC2_IOWIDTH_8:
      mode |= 0 << 0;
      break;
    case ADMXRC2_IOWIDTH_16:
      mode |= 1 << 0;
      break;
    case ADMXRC2_IOWIDTH_32:
      mode |= 3 << 0;
      break;
    default:
      return 0xffffffffU;
    }
    if (waitstates > 0xf) {
      return 0xffffffffU;
    }
    mode |= waitstates << 2;
		
    if (misc & 0xffff263fU) {
      return 0xffffffffU;
    }
    mode |= misc;
    break;
		
  case ADMXRC2_BOARD_ADMXRC2_PRO_LITE:
  case ADMXRC2_BOARD_ADMXRC2_PRO:
    switch (buswidth) {
    case ADMXRC2_IOWIDTH_32:
      mode |= 2 << 0;
      break;
    case ADMXRC2_IOWIDTH_64:
      mode |= 3 << 0;
      break;
    default:
      return 0xffffffffU;
    }
    if (waitstates > 0xf) {
      return 0xffffffffU;
    }
    mode |= waitstates << 2;
		
    if (misc & 0xffff263fU) {
      return 0xffffffffU;
    }
    mode |= misc;
    break;

  default:
    break;
  }
	
  return mode;
}

ADMXRC2_STATUS 
ADMXRC2_CancelWaitForInterrupt(
  ADMXRC2_HANDLE card)
{
  static char         fn[] = "ADMXRC2_CancelWaitForInterrupt";
  ADMXRC2_STATUS      status;

  _ADMXRC2_IoCtlNoWait(card,
		       ADMXRC_IOCTL_CANCELWAITFORINTERRUPT,
		       NULL,
		       0,
		       NULL,
		       0,
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_CloseCard(
  ADMXRC2_HANDLE card)
{
  static char        fn[] = "ADMXRC2_CloseCard";
  ADMXRC2_STATUS     status;

  status = _ADMXRC2_Close(card);
  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_ConfigureFromBuffer(
  ADMXRC2_HANDLE card,
  const void*    buffer,
  unsigned long  length)
{
  static char    fn[] = "ADMXRC2_ConfigureFromBuffer";
  ADMXRC2_STATUS status;

  _ADMXRC2_IoCtlNoWaitMass(card, 
			   (unsigned long) ADMXRC_IOCTL_LOADFPGA,
			   NULL,
			   0,
			   NULL,
			   0,
			   (void*) buffer,
			   length,
			   ADMXRC_IOCTLF_MASS_WRITE,
			   &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_ConfigureFromBufferDMA(
  ADMXRC2_HANDLE card,
  const void*    buffer,
  unsigned long  length,
  unsigned int   dmachan,
  void*          event)
{
  static char               fn[] = "ADMXRC2_ConfigureFromBufferDMA";
  ADMXRC2_STATUS            status;
  ADMXRC_IOCTLS_LOADFPGADMA ioctl;

  ioctl.in.channel = dmachan;
  ioctl.in.flags = 0;
  ioctl.in.buffer = (void*) buffer;
  ioctl.in.length = length;
  _ADMXRC2_IoCtlWait(card,
		     ADMXRC_IOCTL_LOADFPGADMA,
		     &ioctl.in,
		     sizeof(ioctl.in),
		     NULL,
		     0,
		     event,
		     &status);
  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_ConfigureFromFile( 
  ADMXRC2_HANDLE card,
  const char*    filename)
{
  static char    fn[] = "ADMXRC2_ConfigureFromFile";
  unsigned long  size;
  ADMXRC2_IMAGE  image = NULL;
  ADMXRC2_STATUS status;

  status = ADMXRC2_LoadBitstream(card, filename, &image, &size);
  if (status == ADMXRC2_SUCCESS) {
    _ADMXRC2_IoCtlNoWaitMass(card,
			     (unsigned long) ADMXRC_IOCTL_LOADFPGA,
			     NULL,
			     0,
			     NULL,
			     0,
			     image,
			     size,
			     ADMXRC_IOCTLF_MASS_WRITE,
			     &status);
  }
	
  if (image != NULL) {
    ADMXRC2_UnloadBitstream(image);
  }

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_ConfigureFromFileDMA( 
  ADMXRC2_HANDLE card,
  const char*    filename,
  unsigned int   dmachan,
  void*          event)
{
  static char               fn[] = "ADMXRC2_ConfigureFromFileDMA";
  unsigned long             size;
  ADMXRC2_IMAGE             image = NULL;
  ADMXRC2_CARD_INFO         info;
  ADMXRC2_STATUS            status;
  ADMXRC_IOCTLS_LOADFPGADMA ioctl;

  status = ADMXRC2_GetCardInfo(card, &info);
  if (status != ADMXRC2_SUCCESS) {
    goto done;
  }

  status = ADMXRC2_LoadBitstream(card, filename, &image, &size);
  if (status != ADMXRC2_SUCCESS) {
    goto done;
  }

  ioctl.in.channel = dmachan;
  ioctl.in.flags = 0;
  ioctl.in.buffer = (void*) image;
  ioctl.in.length = size;
  _ADMXRC2_IoCtlWait(card,
		     ADMXRC_IOCTL_LOADFPGADMA,
		     &ioctl.in,
		     sizeof(ioctl.in),
		     NULL,
		     0,
		     event,
		     &status);
 done:
  if (image != NULL) {
    ADMXRC2_UnloadBitstream(image);
  }
  
  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_DoDMA(
  ADMXRC2_HANDLE  card,
  ADMXRC2_DMADESC dmadesc,
  unsigned long   offset,
  unsigned long   length,
  uint32_t        local,
  ADMXRC2_DMADIR  direction,
  unsigned int    channel,
  uint32_t        mode,
  uint32_t        flags,
  unsigned long*  timeout,
  void*           event)
{
  static char         fn[] = "ADMXRC2_DoDMA";
  ADMXRC_IOCTLS_DODMA ioctl;
  ADMXRC2_STATUS      status;

  ioctl.in.dmadesc = dmadesc;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = local;
  ioctl.in.direction = direction;
  ioctl.in.channel = channel;
  ioctl.in.mode = mode;
  ioctl.in.flags = flags;

  if (timeout) {
    ioctl.in.timeout = *timeout;
  } else {
    ioctl.in.timeout = 0;
  }

  _ADMXRC2_IoCtlWait(card,
		     ADMXRC_IOCTL_DODMA,
		     &ioctl.in,
		     sizeof(ioctl.in),
		     NULL,
		     0,
		     event,
		     &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_DoDMA2D(
  ADMXRC2_HANDLE  card,
  ADMXRC2_DMADESC dmadesc,
  unsigned long   offset,
  unsigned long   length,
  uint32_t        local,
  unsigned long   width,
  unsigned long   skip,
  unsigned long   local_skip,
  ADMXRC2_DMADIR  direction,
  unsigned int    channel,
  uint32_t        mode,
  uint32_t        flags,
  unsigned long*  timeout,
  void*           event)
{
  static char           fn[] = "ADMXRC2_DoDMA2D";
  ADMXRC_IOCTLS_DODMA2D ioctl;
  ADMXRC2_STATUS        status;

  ioctl.in.dmadesc = dmadesc;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.local = local;
  ioctl.in.width = width;
  ioctl.in.skip = skip;
  ioctl.in.local_skip = local_skip;
  ioctl.in.direction = direction;
  ioctl.in.channel = channel;
  ioctl.in.mode = mode;
  ioctl.in.flags = flags;

  if (timeout) {
    ioctl.in.timeout = *timeout;
  } else {
    ioctl.in.timeout = 0;
  }

  _ADMXRC2_IoCtlWait(card,
		     ADMXRC_IOCTL_DODMA2D,
		     &ioctl.in,
		     sizeof(ioctl.in),
		     NULL,
		     0,
		     event,
		     &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_DoDMAImmediate(
  ADMXRC2_HANDLE card,
  void*          buffer,
  unsigned long  length,
  uint32_t       local,
  ADMXRC2_DMADIR direction,
  unsigned int   channel,
  uint32_t       mode,
  uint32_t       flags,
  unsigned long* timeout,
  void*          event)
{
  static char                  fn[] = "ADMXRC2_DoDMAImmediate";
  ADMXRC_IOCTLS_DODMAIMMEDIATE ioctl;
  ADMXRC2_STATUS               status;

  ioctl.in.buffer = buffer;
  ioctl.in.length = length;
  ioctl.in.local = local;
  ioctl.in.direction = direction;
  ioctl.in.channel = channel;
  ioctl.in.mode = mode;
  ioctl.in.flags = flags;
  if (timeout) {
    ioctl.in.timeout = *timeout;
  } else {
    ioctl.in.timeout = 0;
  }

  _ADMXRC2_IoCtlWait(card,
		     ADMXRC_IOCTL_DODMAIMMEDIATE,
		     &ioctl.in,
		     sizeof(ioctl.in),
		     NULL,
		     0,
		     event,
		     &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_DoDMAImmediate2D(
  ADMXRC2_HANDLE card,
  void*          buffer,
  unsigned long  length,
  uint32_t       local,
  unsigned long  width,
  unsigned long  skip,
  unsigned long  local_skip,
  ADMXRC2_DMADIR direction,
  unsigned int   channel,
  uint32_t       mode,
  uint32_t       flags,
  unsigned long* timeout,
  void*          event)
{
  static char                    fn[] = "ADMXRC2_DoDMAImmediate2D";
  ADMXRC_IOCTLS_DODMAIMMEDIATE2D ioctl;
  ADMXRC2_STATUS                 status;

  ioctl.in.buffer = buffer;
  ioctl.in.length = length;
  ioctl.in.local = local;
  ioctl.in.width = width;
  ioctl.in.skip = skip;
  ioctl.in.local_skip = local_skip;
  ioctl.in.direction = direction;
  ioctl.in.channel = channel;
  ioctl.in.mode = mode;
  ioctl.in.flags = flags;
  if (timeout) {
    ioctl.in.timeout = *timeout;
  } else {
    ioctl.in.timeout = 0;
  }

  _ADMXRC2_IoCtlWait(card,
		     ADMXRC_IOCTL_DODMAIMMEDIATE2D,
		     &ioctl.in,
		     sizeof(ioctl.in),
		     NULL,
		     0,
		     event,
		     &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_EnumerateCards(
  ADMXRC2_ENUM_FUNCTION func)
{
  static char     fn[] = "ADMXRC2_EnumerateCards";
  ADMXRC2_HANDLE  card;
  int	          i, n;
  ADMXRC2_STATUS  status = ADMXRC2_SUCCESS;

  n = 0;
  for (i = 0; i < 16; i++) {
    status = _ADMXRC2_OpenByIndex(i, &card);
    if (status == ADMXRC2_SUCCESS) {
      /* Call the application-supplied callback function */
      func(i, card);
      close(card);
      n++;
    }
  }
  
  if (!n) {
    status = ADMXRC2_CARD_NOT_FOUND;
  }

  if (status != ADMXRC2_SUCCESS) {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  } else {
    return status;
  }
}

ADMXRC2_STATUS
ADMXRC2_EraseFlash(
  ADMXRC2_HANDLE card,
  unsigned long	 start,
  unsigned long	 length,
  const void*    data)
{
  static char              fn[] = "ADMXRC2_EraseFlash";
  ADMXRC_IOCTLS_ERASEFLASH ioctl;
  ADMXRC2_STATUS           status;

  ioctl.in.signature1 = ADMXRC_IOCTLS_ERASEFLASH_SIGNATURE1;
  ioctl.in.start = start;
  ioctl.in.length = length;
  ioctl.in.signature2 = ADMXRC_IOCTLS_ERASEFLASH_SIGNATURE2;
  _ADMXRC2_IoCtlNoWait(card,
		       ADMXRC_IOCTL_ERASEFLASH,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       NULL,
		       0,
		       &status);
  
  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

void
ADMXRC2_Free(
  void* ptr)
{
  _ADMXRC2_PageFree(ptr);
}

ADMXRC2_STATUS  
ADMXRC2_GetBankInfo( 
  ADMXRC2_HANDLE     card,
  unsigned int       banknum,
  ADMXRC2_BANK_INFO* info)
{
  ADMXRC_IOCTLS_GETBANKINFO ioctl;
  static char               fn[] = "ADMXRC2_GetBankInfo";
  ADMXRC2_STATUS            status;

  ioctl.in.index = banknum;
  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_GETBANKINFO,
		       &ioctl,
		       sizeof(ioctl.in),
		       &ioctl,
		       sizeof(ioctl.out),
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    memcpy(info, &ioctl.out.info, sizeof(ADMXRC2_BANK_INFO));
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_GetCardInfo( 
  ADMXRC2_HANDLE     card,
  ADMXRC2_CARD_INFO* info)
{
  ADMXRC_IOCTLS_GETCARDINFO2 ioctl;
  static char                fn[] = "ADMXRC2_GetCardInfo";
  ADMXRC2_STATUS             status;

  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_GETCARDINFO2,
		       NULL,
		       0,
		       &ioctl,
		       sizeof(ioctl.out),
		       &status);
  if (status == ADMXRC2_SUCCESS) {
    memcpy(info, &ioctl.out.info, sizeof(ADMXRC2_CARD_INFO));
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_GetFlashInfo( 
  ADMXRC2_HANDLE      card,
  ADMXRC2_FLASH_INFO* info)
{
  static char                fn[] = "ADMXRC2_GetFlashInfo";
  ADMXRC_IOCTLS_GETFLASHINFO ioctl;
  ADMXRC2_STATUS             status;

  _ADMXRC2_IoCtlNoWait(card,
		       ADMXRC_IOCTL_GETFLASHINFO,
		       NULL,
		       0,
		       &ioctl.out,
		       sizeof(ioctl.out),
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    memcpy(info, &ioctl.out.info, sizeof(*info));
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_GetFlashBlockInfo( 
  ADMXRC2_HANDLE           card,
  unsigned long            location,
  ADMXRC2_FLASHBLOCK_INFO* info)
{
  static char                     fn[] = "ADMXRC2_GetFlashBlockInfo";
  ADMXRC_IOCTLS_GETFLASHBLOCKINFO ioctl;
  ADMXRC2_STATUS                  status;
  
  ioctl.in.location = location;
  _ADMXRC2_IoCtlNoWait(card,
		       ADMXRC_IOCTL_GETFLASHBLOCKINFO,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       &ioctl.out,
		       sizeof(ioctl.out),
		       &status);
		       
  if (status == ADMXRC2_SUCCESS) {
    memcpy(info, &ioctl.out.info, sizeof(*info));
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_GetSpaceConfig(
  ADMXRC2_HANDLE card,
  unsigned int   space_index,
  uint32_t*      flags)
{
  static char                  fn[] = "ADMXRC2_GetSpaceConfig";
  ADMXRC_IOCTLS_GETSPACECONFIG ioctl;
  ADMXRC2_STATUS               status;

  ioctl.in.space_index = space_index;
  _ADMXRC2_IoCtlNoWait(card,
		       ADMXRC_IOCTL_GETSPACECONFIG,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       &ioctl.in,
		       sizeof(ioctl.out),
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    *flags = ioctl.out.flags;
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_GetSpaceInfo( 
  ADMXRC2_HANDLE      card,
  unsigned int        space_index,
  ADMXRC2_SPACE_INFO* info)
{
  ADMXRC_IOCTLS_GETSPACEINFO ioctl;
  static char                fn[] = "ADMXRC2_GetSpaceInfo";
  ADMXRC2_STATUS             status;

  ioctl.in.index = (unsigned long) space_index;
  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_GETSPACEINFO,
		       &ioctl,
		       sizeof(ioctl.in),
		       &ioctl,
		       sizeof(ioctl.out),
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    memcpy(info, &ioctl.out.info, sizeof(ADMXRC2_SPACE_INFO));
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

const char*
ADMXRC2_GetStatusString(
  ADMXRC2_STATUS Status)
{
  int	i;
	
  i = 0;
  while (statusMessage[i].code != ADMXRC2_LASTERROR) {
    if (statusMessage[i].code == Status) {
      return statusMessage[i].description;
    }
    i++;
  }

  return "unknown ADMXRC2_STATUS code";
}

ADMXRC2_STATUS 
ADMXRC2_GetVersionInfo(
  ADMXRC2_HANDLE        card,
  ADMXRC2_VERSION_INFO* info)
{
  ADMXRC_IOCTLS_GETVERSIONINFO ioctl;
  static char                  fn[] = "ADMXRC2_GetVersionInfo";
  ADMXRC2_STATUS               status;

  _ADMXRC2_IoCtlNoWait(card, 
		       (unsigned long) ADMXRC_IOCTL_GETVERSIONINFO,
		       NULL,
		       0,
		       &ioctl.out,
		       sizeof(ioctl.out),
		       &status);
  if (status == ADMXRC2_SUCCESS) {
    info->DriverMajor = ioctl.out.major;
    info->DriverMinor = ioctl.out.minor;
    info->APIMajor = ADMXRC_API_MAJORVER;
    info->APIMinor = ADMXRC_API_MINORVER;
    return ADMXRC2_SUCCESS;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_InstallErrorHandler(
  ADMXRC2_ERROR_HANDLER handler)
{
  _ADMXRC2_errorHandler = handler;

  return ADMXRC2_SUCCESS;
}


ADMXRC2_STATUS
ADMXRC2_LoadBitstream(
  ADMXRC2_HANDLE  card,
  const char*     filename,
  ADMXRC2_IMAGE*  image,
  unsigned long*  image_size)
{
  ADMXRC2_CARD_INFO info;
  ADMXRC2_STATUS    status;
  FILE*             file;
  long              size, pos;
  void*             buf = NULL;
  long              numread;

  status = ADMXRC2_GetCardInfo(card, &info);
  if (status != ADMXRC2_SUCCESS) {
    goto done;
  }

  file = fopen(filename, "r");
  if (file == NULL) {
    status = ADMXRC2_FILE_NOT_FOUND;
    goto done;
  }

  status = SelectMapOffset(&info, file, &size, &pos);
  if (status != ADMXRC2_SUCCESS) {
    fclose(file);
    goto done;
  }

  buf = ADMXRC2_Malloc(size);
  if (buf == NULL) {
    status = ADMXRC2_NO_MEMORY;
    fclose(file);
    goto done;
  }
  fseek(file, pos, SEEK_SET);
  numread = fread(buf, 1, size, file);
  if (!numread) {
    status = ADMXRC2_INVALID_FILE;
    ADMXRC2_Free(buf);
    fclose(file);
    goto done;
  }
  
  switch (info.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADMXRC2_PRO_LITE:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADMXRC2_PRO:
    ReverseBytes(buf, size);
    break;
  default:
    break;
  }
  
  fclose(file);
  *image = (ADMXRC2_IMAGE) buf;
  *image_size = size;
  
 done:
  return status;
}

void*
ADMXRC2_Malloc(
  unsigned long nbytes)
{
  return _ADMXRC2_PageMalloc(nbytes);
}

ADMXRC2_STATUS
ADMXRC2_MapDirectMaster(
  ADMXRC2_HANDLE     card,
  ADMXRC2_DMADESC    dmadesc,
  unsigned long      offset,
  unsigned long      length,
  ADMXRC2_BUFFERMAP* map)
{
  static char                   fn[] = "ADMXRC2_MapDirectMaster";
  ADMXRC_IOCTLS_MAPDIRECTMASTER ioctl;
  ADMXRC2_STATUS                status;

  ioctl.in.dmadesc = dmadesc;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.maxpage = map->MaxPages;

  _ADMXRC2_IoCtlNoWaitMass(card,
			   (unsigned long) ADMXRC_IOCTL_MAPDIRECTMASTER,
			   &ioctl.in,
			   sizeof(ioctl.in),
			   &ioctl.out,
			   sizeof(ioctl.out),
			   map->PagesPci,
			   sizeof(unsigned long) * map->MaxPages,
			   ADMXRC_IOCTLF_MASS_READ,
			   &status);

  if (status == ADMXRC2_SUCCESS) {
    map->PageLength = ioctl.out.pagelen;
    map->PageBits = ioctl.out.pagebits;
    map->PagesSpanned = ioctl.out.pagespan;
    map->BytesSpanned = ioctl.out.bytespan;
    map->InitOffset = ioctl.out.initoffs;
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_OpenCardByIndex(
  unsigned int    index,
  ADMXRC2_HANDLE* card)
{
  static char                  fn[] = "ADMXRC2_OpenCardByIndex";
  ADMXRC2_HANDLE               handle;
  ADMXRC2_STATUS               status;
  ADMXRC_IOCTLS_GETVERSIONINFO ioctl;

  status = _ADMXRC2_OpenByIndex(index, &handle);
  if (status != ADMXRC2_SUCCESS) {
    goto done;
  }

  _ADMXRC2_IoCtlNoWait(handle,
		       (unsigned long) ADMXRC_IOCTL_OPENCARDBYINDEX,
		       NULL,
		       0,
		       NULL,
		       0,
		       &status);
  if (status != ADMXRC2_SUCCESS) {
    _ADMXRC2_Close(handle);
    goto done;
  }

  _ADMXRC2_IoCtlNoWait(handle, 
		       (unsigned long) ADMXRC_IOCTL_GETVERSIONINFO,
		       NULL,
		       0,
		       &ioctl.out,
		       sizeof(ioctl.out),
		       &status);
  if (status != ADMXRC2_SUCCESS ||
      (ioctl.out.major < ADMXRC_DVR_MAJORVER) ||
      (ioctl.out.major == ADMXRC_DVR_MAJORVER && ioctl.out.minor < ADMXRC_DVR_MINORVER)) {
    fprintf(stderr,"The ADM-XRC device driver in this sytem is not correctly installed.\n\n"
	    "Please reinstall the ADM-XRC device driver version %d.%d or later.\n",
	    ADMXRC_DVR_MAJORVER, ADMXRC_DVR_MINORVER);
    _ADMXRC2_Close(handle);
    status = ADMXRC2_BAD_DRIVER;
    goto done;
  }

  status = _ADMXRC2_MapToUserSpace(handle);
  if (status != ADMXRC2_SUCCESS) {
    close(handle);
    goto done;
  }

  *card = handle;

 done:
  if (status != ADMXRC2_SUCCESS) {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  } else {
    return status;
  }
}

ADMXRC2_STATUS 
ADMXRC2_OpenCard(
  ADMXRC2_CARDID  cardID,
  ADMXRC2_HANDLE* card)
{
  static char        fn[] = "ADMXRC2_OpenCard";
  ADMXRC2_HANDLE     handle;
  int                i;
  ADMXRC2_STATUS     status;
  
  for (i = 0; i < 16; i++) {
    status = _ADMXRC2_OpenByIndex(i, &handle);
    if (status == ADMXRC2_SUCCESS) {
      ADMXRC_IOCTLS_OPENCARD2 ioctl;

      ioctl.in.card_id = cardID;
      _ADMXRC2_IoCtlNoWait(handle,
			   (unsigned long) ADMXRC_IOCTL_OPENCARD2,
			   &ioctl,
			   sizeof(ioctl.in),
			   NULL,
			   0,
			   &status);
      if (status == ADMXRC2_SUCCESS) {
	ADMXRC_IOCTLS_GETVERSIONINFO ioctl;
	   
        _ADMXRC2_IoCtlNoWait(handle, 
			     (unsigned long) ADMXRC_IOCTL_GETVERSIONINFO,
			     NULL,
			     0,
			     &ioctl.out,
			     sizeof(ioctl.out),
			     &status);
	if (status != ADMXRC2_SUCCESS ||
	    (ioctl.out.major < ADMXRC_DVR_MAJORVER) ||
	    (ioctl.out.major == ADMXRC_DVR_MAJORVER && ioctl.out.minor < ADMXRC_DVR_MINORVER)) {
	  fprintf(stderr, "The ADM-XRC device driver in this sytem is not correctly installed.\n\n"
		  "Please reinstall the ADM-XRC device driver version %d.%d or later.\n",
		  ADMXRC_DVR_MAJORVER, ADMXRC_DVR_MINORVER);
	  _ADMXRC2_Close(handle);  
	  status = ADMXRC2_BAD_DRIVER;
	  goto done;
	}

	status = _ADMXRC2_MapToUserSpace(handle);
	if (status != ADMXRC2_SUCCESS) {
	  _ADMXRC2_Close(handle);
	  goto done;
	}

	*card = handle;
	return ADMXRC2_SUCCESS;
      } else {
	/* Not the card we're looking for, so close it */
	_ADMXRC2_Close(handle);
      }
    }
  }
 
  status = ADMXRC2_CARD_NOT_FOUND;

 done:  
  if (_ADMXRC2_errorHandler != NULL) {
    return _ADMXRC2_handleError(fn, status);   
  } else {
    return status;
  }
}

ADMXRC2_STATUS 
ADMXRC2_ProgramBytes(
  ADMXRC2_HANDLE       card,
  const uint32_t*      program,
  unsigned long        length)
{
  static char                fn[] = "ADMXRC2_ProgramBytes";
  ADMXRC2_STATUS             status;

  _ADMXRC2_IoCtlNoWaitMass(card, 
			   (unsigned long) ADMXRC_IOCTL_PROGRAMBYTES,
			   NULL,
			   0,
			   NULL,
			   0,
			   (void*) program,
			   length * 2 * sizeof(unsigned int),
			   ADMXRC_IOCTLF_MASS_READWRITE,
			   &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_Read(
  ADMXRC2_HANDLE  card,
  ADMXRC2_IOWIDTH size,
  uint32_t        flags,
  uint32_t        local,
  void*           data,
  unsigned long   len)
{
  static char        fn[] = "ADMXRC2_Read";
  ADMXRC_IOCTLS_READ ioctl;
  ADMXRC2_STATUS     status;

  ioctl.in.width = size;
  ioctl.in.local = local;
  ioctl.in.flags = flags;
  _ADMXRC2_IoCtlNoWaitMass(card,
			   (unsigned long) ADMXRC_IOCTL_READ,
			   &ioctl,
			   sizeof(ioctl.in),
			   NULL,
			   0,
			   data,
			   len,
			   ADMXRC_IOCTLF_MASS_READ,
			   &status);
  
  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_ReadConfig(
  ADMXRC2_HANDLE card,
  unsigned long  index,
  uint32_t*      value)
{
  ADMXRC_IOCTLS_READCONFIG ioctl;
  static char              fn[] = "ADMXRC2_ReadConfig";
  ADMXRC2_STATUS           status;

  ioctl.in.index = (unsigned long) index;
  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_READCONFIG,
		       &ioctl, 
		       sizeof(ioctl.in),
		       &ioctl,
		       sizeof(ioctl.out),
		       &status);
  if (status == ADMXRC2_SUCCESS) {
    *value = ioctl.out.value;
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_ReadFlash(
  ADMXRC2_HANDLE card,
  unsigned long	 start,
  unsigned long	 length,
  void*          buffer)
{
  static char             fn[] = "ADMXRC2_ReadFlash";
  ADMXRC_IOCTLS_READFLASH ioctl;
  ADMXRC2_STATUS          status;
  
  ioctl.in.start = start;
  ioctl.in.length = length;
  _ADMXRC2_IoCtlNoWaitMass(card,
		       ADMXRC_IOCTL_READFLASH,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       NULL,
		       0,
		       buffer,
		       length,
		       ADMXRC_IOCTLF_MASS_READ,
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_SetClockRate(
  ADMXRC2_HANDLE card,
  unsigned int   clock,
  double         freq,
  double*        actual)
{
  static char             fn[] = "ADMXRC2_SetClockRate";
  ADMXRC2_STATUS          status;
  ADMXRC_IOCTLS_SETCLOCK2 ioctl;

  ioctl.in.clock_index = (unsigned long) clock;
  ioctl.in.freq = (unsigned long) freq;
  _ADMXRC2_IoCtlNoWait(card, 
		       (unsigned long) ADMXRC_IOCTL_SETCLOCK2,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       &ioctl.out,
		       sizeof(ioctl.out),
		       &status);
  if (status == ADMXRC2_SUCCESS) {
    if (actual != NULL) {
	*actual = (double) ioctl.out.actual;
    }
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_SetSpaceConfig(
  ADMXRC2_HANDLE card,
  unsigned int   space_index,
  uint32_t       flags)
{
  static char                  fn[] = "ADMXRC2_SetSpaceConfig";
  ADMXRC_IOCTLS_SETSPACECONFIG ioctl;
  ADMXRC2_STATUS               status;

  ioctl.in.space_index = space_index;
  ioctl.in.flags = flags;
  _ADMXRC2_IoCtlNoWait(card,
		       ADMXRC_IOCTL_SETSPACECONFIG,
		       &ioctl,
		       sizeof(ioctl),
		       NULL,
		       0,
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_SetupDMA(
  ADMXRC2_HANDLE   card,
  void*            buffer,
  unsigned long    size,
  uint32_t         flags,
  ADMXRC2_DMADESC* dmadesc)
{
  static char            fn[] = "ADMXRC2_SetupDMA";
  ADMXRC_IOCTLS_SETUPDMA ioctl;
  ADMXRC2_STATUS         status;

  ioctl.in.buffer = buffer;
  ioctl.in.size = size;
  ioctl.in.flags = flags;

  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_SETUPDMA,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       &ioctl.out,
		       sizeof(ioctl.out),
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    *dmadesc = (ADMXRC2_DMADESC) ioctl.out.dmadesc;
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_SyncDirectMaster(
  ADMXRC2_HANDLE    card,
  ADMXRC2_DMADESC   dmadesc,
  unsigned long     offset,
  unsigned long     length,
  ADMXRC2_SYNCMODE  syncmode)
{
  ADMXRC_IOCTLS_SYNCDIRECTMASTER ioctl;
  static char                    fn[] = "ADMXRC2_SyncDirectMaster";
  ADMXRC2_STATUS                 status;

  ioctl.in.dmadesc = dmadesc;
  ioctl.in.offset = offset;
  ioctl.in.length = length;
  ioctl.in.syncmode = syncmode;

  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_SYNCDIRECTMASTER,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       NULL,
		       0,
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }  
}

ADMXRC2_STATUS
ADMXRC2_UnloadBitstream(
  ADMXRC2_IMAGE image)
{
  if (image != NULL) {
    ADMXRC2_Free(image);
  }

  return ADMXRC2_SUCCESS;
}

ADMXRC2_STATUS 
ADMXRC2_UnsetupDMA(
  ADMXRC2_HANDLE  card,
  ADMXRC2_DMADESC dmadesc)
{
  static char              fn[] = "ADMXRC2_UnsetupDMA";
  ADMXRC_IOCTLS_UNSETUPDMA ioctl;
  ADMXRC2_STATUS           status;

  ioctl.in.dmadesc = dmadesc;
  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_UNSETUPDMA,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       NULL,
		       0,
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_WaitForInterrupt(
  ADMXRC2_HANDLE  card,
  unsigned long*  timeout_us,
  unsigned long   flags,
  void*           event)
{
  static char                    fn[] = "ADMXRC2_WaitForInterrupt";
  ADMXRC_IOCTLS_WAITFORINTERRUPT ioctl;
  ADMXRC2_STATUS                 status;

  ioctl.in.flags = 0;
  if (timeout_us != NULL) {
    ioctl.in.timeout_us = *timeout_us;
    ioctl.in.flags |= ADMXRC_INTWAIT_TIMEOUT;
  }

  _ADMXRC2_IoCtlWaitMass(card,
			 (unsigned long) ADMXRC_IOCTL_WAITFORINTERRUPT,
			 &ioctl.in,
			 sizeof(ioctl.in),
			 &ioctl.out,
			 sizeof(ioctl.out),
			 NULL,
			 0,
			 ADMXRC_IOCTLF_MASS_WRITE,
			 event,
			 &status);

  if (status == ADMXRC2_SUCCESS) {
    if (timeout_us != NULL) {
      *timeout_us = ioctl.out.remaining_us;
    }
    if (ioctl.out.expired) {
      return ADMXRC2_TIMEOUT;
    } else {
      return ADMXRC2_SUCCESS;
    }
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_Write(
  ADMXRC2_HANDLE  card,
  ADMXRC2_IOWIDTH size,
  uint32_t        flags,
  uint32_t        local,
  void*           data,
  unsigned long   len)
{
  static char         fn[] = "ADMXRC2_Write";
  ADMXRC_IOCTLS_WRITE ioctl;
  ADMXRC2_STATUS      status;

  ioctl.in.width = size;
  ioctl.in.local = local;
  ioctl.in.flags = flags;
  _ADMXRC2_IoCtlNoWaitMass(card,
			   (unsigned long) ADMXRC_IOCTL_WRITE,
			   &ioctl,
			   sizeof(ioctl.in),
			   NULL,
			   0,
			   data,
			   len,
			   ADMXRC_IOCTLF_MASS_WRITE,
			   &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS  
ADMXRC2_WriteConfig(
  ADMXRC2_HANDLE card,
  unsigned long  index,
  uint32_t       value)
{
  static char               fn[] = "ADMXRC2_WriteConfig";
  ADMXRC_IOCTLS_WRITECONFIG ioctl;
  ADMXRC2_STATUS            status;

  ioctl.in.signature1 = ADMXRC_IOCTLS_WRITECONFIG_SIGNATURE1;
  ioctl.in.index = index;
  ioctl.in.value = value;
  ioctl.in.signature2 = ADMXRC_IOCTLS_WRITECONFIG_SIGNATURE2;
  _ADMXRC2_IoCtlNoWait(card,
		       (unsigned long) ADMXRC_IOCTL_WRITECONFIG,
		       &ioctl,
		       sizeof(ioctl.in),
		       NULL,
		       0,
		       &status);
  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_WriteFlash(
  ADMXRC2_HANDLE card,
  unsigned long  start,
  unsigned long  length,
  const void*    data)
{
  static char              fn[] = "ADMXRC2_WriteFlash";
  ADMXRC_IOCTLS_WRITEFLASH ioctl;
  ADMXRC2_STATUS           status;

  ioctl.in.signature1 = ADMXRC_IOCTLS_WRITEFLASH_SIGNATURE1;
  ioctl.in.start = start;
  ioctl.in.length = length;
  ioctl.in.signature2 = ADMXRC_IOCTLS_WRITEFLASH_SIGNATURE2;
  _ADMXRC2_IoCtlNoWaitMass(card,
		       ADMXRC_IOCTL_WRITEFLASH,
		       &ioctl.in,
		       sizeof(ioctl.in),
		       NULL,
		       0,
		       (void*) data,
		       length,
		       ADMXRC_IOCTLF_MASS_WRITE,
		       &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS
ADMXRC2_RegisterInterruptEvent(
  ADMXRC2_HANDLE card,
  os_event       event)
{
  static char                     fn[] = "ADMXRC2_RegisterInterruptEvent";
  ADMXRC_IOCTLS_REGINTERRUPTEVENT ioctl;
  ADMXRC2_STATUS                  status;

  ioctl.in.event = event;

  _ADMXRC2_IoCtlNoWait(card, 
		      (unsigned int) ADMXRC_IOCTL_REGINTERRUPTEVENT,
		      &ioctl,
		      sizeof(ioctl),
		      NULL,
		      0,
		      &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

ADMXRC2_STATUS 
ADMXRC2_UnregisterInterruptEvent(
  ADMXRC2_HANDLE card,
  os_event       event)
{
  static char                       fn[] = "ADMXRC2_UnregisterInterruptEvent";
  ADMXRC_IOCTLS_UNREGINTERRUPTEVENT ioctl;
  ADMXRC2_STATUS                    status;

  ioctl.in.event = event;

  _ADMXRC2_IoCtlNoWait(card, 
		      (unsigned int) ADMXRC_IOCTL_UNREGINTERRUPTEVENT,
		      &ioctl,
		      sizeof(ioctl),
		      NULL,
		      0,
		      &status);

  if (status == ADMXRC2_SUCCESS) {
    return status;
  } else {
    if (_ADMXRC2_errorHandler != NULL) {
      return _ADMXRC2_handleError(fn, status);
    } else {
      return status;
    }
  }
}

